import math
import warnings
import random
import numbers
import numpy as np
from PIL import Image, ImageFilter
from collections.abc import Sequence

import torch
import torchvision.transforms.functional as TF

_pil_interpolation_to_str = {
    Image.NEAREST: 'PIL.Image.NEAREST',
    Image.BILINEAR: 'PIL.Image.BILINEAR',
    Image.BICUBIC: 'PIL.Image.BICUBIC',
    Image.LANCZOS: 'PIL.Image.LANCZOS',
    Image.HAMMING: 'PIL.Image.HAMMING',
    Image.BOX: 'PIL.Image.BOX',
}


def _get_image_size(img):
    if TF._is_pil_image(img):
        return img.size
    elif isinstance(img, torch.Tensor) and img.dim() > 2:
        return img.shape[-2:][::-1]
    else:
        raise TypeError("Unexpected type {}".format(type(img)))


class RandomHorizontalFlip(object):
    """Horizontal flip the given PIL Image randomly with a given probability.

    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """
    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, img, mask):
        """
        Args:
            img (PIL Image): Image to be flipped.

        Returns:
            PIL Image: Randomly flipped image.
        """
        if random.random() < self.p:
            img = TF.hflip(img)
            mask = TF.hflip(mask)
        return img, mask

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class RandomVerticalFlip(object):
    """Vertical flip the given PIL Image randomly with a given probability.

    Args:
        p (float): probability of the image being flipped. Default value is 0.5
    """
    def __init__(self, p=0.5):
        self.p = p

    def __call__(self, img, mask):
        """
        Args:
            img (PIL Image): Image to be flipped.

        Returns:
            PIL Image: Randomly flipped image.
        """
        if random.random() < self.p:
            img = TF.vflip(img)
            mask = TF.vflip(mask)
        return img, mask

    def __repr__(self):
        return self.__class__.__name__ + '(p={})'.format(self.p)


class GaussianBlur(object):
    """Gaussian blur augmentation from SimCLR: https://arxiv.org/abs/2002.05709"""
    def __init__(self, sigma=[.1, 2.]):
        self.sigma = sigma

    def __call__(self, x):
        sigma = random.uniform(self.sigma[0], self.sigma[1])
        x = x.filter(ImageFilter.GaussianBlur(radius=sigma))
        return x


class RandomAffine(object):
    """Random affine transformation of the image keeping center invariant

    Args:
        degrees (sequence or float or int): Range of degrees to select from.
            If degrees is a number instead of sequence like (min, max), the range of degrees
            will be (-degrees, +degrees). Set to 0 to deactivate rotations.
        translate (tuple, optional): tuple of maximum absolute fraction for horizontal
            and vertical translations. For example translate=(a, b), then horizontal shift
            is randomly sampled in the range -img_width * a < dx < img_width * a and vertical shift is
            randomly sampled in the range -img_height * b < dy < img_height * b. Will not translate by default.
        scale (tuple, optional): scaling factor interval, e.g (a, b), then scale is
            randomly sampled from the range a <= scale <= b. Will keep original scale by default.
        shear (sequence or float or int, optional): Range of degrees to select from.
            If shear is a number, a shear parallel to the x axis in the range (-shear, +shear)
            will be apllied. Else if shear is a tuple or list of 2 values a shear parallel to the x axis in the
            range (shear[0], shear[1]) will be applied. Else if shear is a tuple or list of 4 values,
            a x-axis shear in (shear[0], shear[1]) and y-axis shear in (shear[2], shear[3]) will be applied.
            Will not apply shear by default
        resample ({PIL.Image.NEAREST, PIL.Image.BILINEAR, PIL.Image.BICUBIC}, optional):
            An optional resampling filter. See `filters`_ for more information.
            If omitted, or if the image has mode "1" or "P", it is set to PIL.Image.NEAREST.
        fillcolor (tuple or int): Optional fill color (Tuple for RGB Image And int for grayscale) for the area
            outside the transform in the output image.(Pillow>=5.0.0)

    .. _filters: https://pillow.readthedocs.io/en/latest/handbook/concepts.html#filters

    """
    def __init__(self,
                 degrees,
                 translate=None,
                 scale=None,
                 shear=None,
                 resample=False,
                 fillcolor=0):
        if isinstance(degrees, numbers.Number):
            if degrees < 0:
                raise ValueError(
                    "If degrees is a single number, it must be positive.")
            self.degrees = (-degrees, degrees)
        else:
            assert isinstance(degrees, (tuple, list)) and len(degrees) == 2, \
                "degrees should be a list or tuple and it must be of length 2."
            self.degrees = degrees

        if translate is not None:
            assert isinstance(translate, (tuple, list)) and len(translate) == 2, \
                "translate should be a list or tuple and it must be of length 2."
            for t in translate:
                if not (0.0 <= t <= 1.0):
                    raise ValueError(
                        "translation values should be between 0 and 1")
        self.translate = translate

        if scale is not None:
            assert isinstance(scale, (tuple, list)) and len(scale) == 2, \
                "scale should be a list or tuple and it must be of length 2."
            for s in scale:
                if s <= 0:
                    raise ValueError("scale values should be positive")
        self.scale = scale

        if shear is not None:
            if isinstance(shear, numbers.Number):
                if shear < 0:
                    raise ValueError(
                        "If shear is a single number, it must be positive.")
                self.shear = (-shear, shear)
            else:
                assert isinstance(shear, (tuple, list)) and \
                       (len(shear) == 2 or len(shear) == 4), \
                       "shear should be a list or tuple and it must be of length 2 or 4."
                # X-Axis shear with [min, max]
                if len(shear) == 2:
                    self.shear = [shear[0], shear[1], 0., 0.]
                elif len(shear) == 4:
                    self.shear = [s for s in shear]
        else:
            self.shear = shear

        self.resample = resample
        self.fillcolor = fillcolor

    @staticmethod
    def get_params(degrees, translate, scale_ranges, shears, img_size):
        """Get parameters for affine transformation

        Returns:
            sequence: params to be passed to the affine transformation
        """
        angle = random.uniform(degrees[0], degrees[1])
        if translate is not None:
            max_dx = translate[0] * img_size[0]
            max_dy = translate[1] * img_size[1]
            translations = (np.round(random.uniform(-max_dx, max_dx)),
                            np.round(random.uniform(-max_dy, max_dy)))
        else:
            translations = (0, 0)

        if scale_ranges is not None:
            scale = random.uniform(scale_ranges[0], scale_ranges[1])
        else:
            scale = 1.0

        if shears is not None:
            if len(shears) == 2:
                shear = [random.uniform(shears[0], shears[1]), 0.]
            elif len(shears) == 4:
                shear = [
                    random.uniform(shears[0], shears[1]),
                    random.uniform(shears[2], shears[3])
                ]
        else:
            shear = 0.0

        return angle, translations, scale, shear

    def __call__(self, img, mask):
        """
            img (PIL Image): Image to be transformed.

        Returns:
            PIL Image: Affine transformed image.
        """
        ret = self.get_params(self.degrees, self.translate, self.scale,
                              self.shear, img.size)
        img = TF.affine(img,
                        *ret,
                        resample=self.resample,
                        fillcolor=self.fillcolor)
        mask = TF.affine(mask, *ret, resample=Image.NEAREST, fillcolor=0)
        return img, mask

    def __repr__(self):
        s = '{name}(degrees={degrees}'
        if self.translate is not None:
            s += ', translate={translate}'
        if self.scale is not None:
            s += ', scale={scale}'
        if self.shear is not None:
            s += ', shear={shear}'
        if self.resample > 0:
            s += ', resample={resample}'
        if self.fillcolor != 0:
            s += ', fillcolor={fillcolor}'
        s += ')'
        d = dict(self.__dict__)
        d['resample'] = _pil_interpolation_to_str[d['resample']]
        return s.format(name=self.__class__.__name__, **d)


class RandomCrop(object):
    """Crop the given PIL Image at a random location.

    Args:
        size (sequence or int): Desired output size of the crop. If size is an
            int instead of sequence like (h, w), a square crop (size, size) is
            made.
        padding (int or sequence, optional): Optional padding on each border
            of the image. Default is None, i.e no padding. If a sequence of length
            4 is provided, it is used to pad left, top, right, bottom borders
            respectively. If a sequence of length 2 is provided, it is used to
            pad left/right, top/bottom borders, respectively.
        pad_if_needed (boolean): It will pad the image if smaller than the
            desired size to avoid raising an exception. Since cropping is done
            after padding, the padding seems to be done at a random offset.
        fill: Pixel fill value for constant fill. Default is 0. If a tuple of
            length 3, it is used to fill R, G, B channels respectively.
            This value is only used when the padding_mode is constant
        padding_mode: Type of padding. Should be: constant, edge, reflect or symmetric. Default is constant.

             - constant: pads with a constant value, this value is specified with fill

             - edge: pads with the last value on the edge of the image

             - reflect: pads with reflection of image (without repeating the last value on the edge)

                padding [1, 2, 3, 4] with 2 elements on both sides in reflect mode
                will result in [3, 2, 1, 2, 3, 4, 3, 2]

             - symmetric: pads with reflection of image (repeating the last value on the edge)

                padding [1, 2, 3, 4] with 2 elements on both sides in symmetric mode
                will result in [2, 1, 1, 2, 3, 4, 4, 3]

    """
    def __init__(self,
                 size,
                 padding=None,
                 pad_if_needed=False,
                 fill=0,
                 padding_mode='constant'):
        if isinstance(size, numbers.Number):
            self.size = (int(size), int(size))
        else:
            self.size = size
        self.padding = padding
        self.pad_if_needed = pad_if_needed
        self.fill = fill
        self.padding_mode = padding_mode

    @staticmethod
    def get_params(img, output_size):
        """Get parameters for ``crop`` for a random crop.

        Args:
            img (PIL Image): Image to be cropped.
            output_size (tuple): Expected output size of the crop.

        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for random crop.
        """
        w, h = _get_image_size(img)
        th, tw = output_size
        if w == tw and h == th:
            return 0, 0, h, w

        i = random.randint(0, h - th)
        j = random.randint(0, w - tw)
        return i, j, th, tw

    def __call__(self, img, mask):
        """
        Args:
            img (PIL Image): Image to be cropped.

        Returns:
            PIL Image: Cropped image.
        """
        # if self.padding is not None:
        #     img = TF.pad(img, self.padding, self.fill, self.padding_mode)
        #
        # # pad the width if needed
        # if self.pad_if_needed and img.size[0] < self.size[1]:
        #     img = TF.pad(img, (self.size[1] - img.size[0], 0), self.fill, self.padding_mode)
        # # pad the height if needed
        # if self.pad_if_needed and img.size[1] < self.size[0]:
        #     img = TF.pad(img, (0, self.size[0] - img.size[1]), self.fill, self.padding_mode)

        i, j, h, w = self.get_params(img, self.size)
        img = TF.crop(img, i, j, h, w)
        mask = TF.crop(mask, i, j, h, w)

        return img, mask

    def __repr__(self):
        return self.__class__.__name__ + '(size={0}, padding={1})'.format(
            self.size, self.padding)


class RandomResizedCrop(object):
    """Crop the given PIL Image to random size and aspect ratio.

    A crop of random size (default: of 0.08 to 1.0) of the original size and a random
    aspect ratio (default: of 3/4 to 4/3) of the original aspect ratio is made. This crop
    is finally resized to given size.
    This is popularly used to train the Inception networks.

    Args:
        size: expected output size of each edge
        scale: range of size of the origin size cropped
        ratio: range of aspect ratio of the origin aspect ratio cropped
        interpolation: Default: PIL.Image.BILINEAR
    """
    def __init__(self,
                 size,
                 scale=(0.08, 1.0),
                 ratio=(3. / 4., 4. / 3.),
                 interpolation=Image.BILINEAR):
        if isinstance(size, (tuple, list)):
            self.size = size
        else:
            self.size = (size, size)
        if (scale[0] > scale[1]) or (ratio[0] > ratio[1]):
            warnings.warn("range should be of kind (min, max)")

        self.interpolation = interpolation
        self.scale = scale
        self.ratio = ratio

    @staticmethod
    def get_params(img, scale, ratio):
        """Get parameters for ``crop`` for a random sized crop.

        Args:
            img (PIL Image): Image to be cropped.
            scale (tuple): range of size of the origin size cropped
            ratio (tuple): range of aspect ratio of the origin aspect ratio cropped

        Returns:
            tuple: params (i, j, h, w) to be passed to ``crop`` for a random
                sized crop.
        """
        width, height = _get_image_size(img)
        area = height * width

        for _ in range(10):
            target_area = random.uniform(*scale) * area
            log_ratio = (math.log(ratio[0]), math.log(ratio[1]))
            aspect_ratio = math.exp(random.uniform(*log_ratio))

            w = int(round(math.sqrt(target_area * aspect_ratio)))
            h = int(round(math.sqrt(target_area / aspect_ratio)))

            if 0 < w <= width and 0 < h <= height:
                i = random.randint(0, height - h)
                j = random.randint(0, width - w)
                return i, j, h, w

        # Fallback to central crop
        in_ratio = float(width) / float(height)
        if (in_ratio < min(ratio)):
            w = width
            h = int(round(w / min(ratio)))
        elif (in_ratio > max(ratio)):
            h = height
            w = int(round(h * max(ratio)))
        else:  # whole image
            w = width
            h = height
        i = (height - h) // 2
        j = (width - w) // 2
        return i, j, h, w

    def __call__(self, img, mask):
        """
        Args:
            img (PIL Image): Image to be cropped and resized.

        Returns:
            PIL Image: Randomly cropped and resized image.
        """
        i, j, h, w = self.get_params(img, self.scale, self.ratio)
        # print(i, j, h, w)
        img = TF.resized_crop(img, i, j, h, w, self.size, self.interpolation)
        mask = TF.resized_crop(mask, i, j, h, w, self.size, Image.NEAREST)
        return img, mask

    def __repr__(self):
        interpolate_str = _pil_interpolation_to_str[self.interpolation]
        format_string = self.__class__.__name__ + '(size={0}'.format(self.size)
        format_string += ', scale={0}'.format(
            tuple(round(s, 4) for s in self.scale))
        format_string += ', ratio={0}'.format(
            tuple(round(r, 4) for r in self.ratio))
        format_string += ', interpolation={0})'.format(interpolate_str)
        return format_string


class ToOnehot(object):
    """To oneshot tensor

    Args:
        max_obj_n (float): Maximum number of the objects
    """
    def __init__(self, max_obj_n, shuffle):
        self.max_obj_n = max_obj_n
        self.shuffle = shuffle

    def __call__(self, mask, obj_list=None):
        """
        Args:
            mask (Mask in Numpy): Mask to be converted.

        Returns:
            Tensor: Converted mask in onehot format.
        """

        new_mask = np.zeros((self.max_obj_n + 1, *mask.shape), np.uint8)

        if not obj_list:
            obj_list = list()
            obj_max = mask.max() + 1
            for i in range(1, obj_max):
                tmp = (mask == i).astype(np.uint8)
                if tmp.max() > 0:
                    obj_list.append(i)

            if self.shuffle:
                random.shuffle(obj_list)
            obj_list = obj_list[:self.max_obj_n]

        for i in range(len(obj_list)):
            new_mask[i + 1] = (mask == obj_list[i]).astype(np.uint8)
        new_mask[0] = 1 - np.sum(new_mask, axis=0)

        return torch.from_numpy(new_mask), obj_list

    def __repr__(self):
        return self.__class__.__name__ + '(max_obj_n={})'.format(
            self.max_obj_n)


class Resize(torch.nn.Module):
    """Resize the input image to the given size.
    The image can be a PIL Image or a torch Tensor, in which case it is expected
    to have [..., H, W] shape, where ... means an arbitrary number of leading dimensions

    Args:
        size (sequence or int): Desired output size. If size is a sequence like
            (h, w), output size will be matched to this. If size is an int,
            smaller edge of the image will be matched to this number.
            i.e, if height > width, then image will be rescaled to
            (size * height / width, size).
            In torchscript mode padding as single int is not supported, use a tuple or
            list of length 1: ``[size, ]``.
        interpolation (int, optional): Desired interpolation enum defined by `filters`_.
            Default is ``PIL.Image.BILINEAR``. If input is Tensor, only ``PIL.Image.NEAREST``, ``PIL.Image.BILINEAR``
            and ``PIL.Image.BICUBIC`` are supported.
    """
    def __init__(self, size, interpolation=Image.BILINEAR):
        super().__init__()
        if not isinstance(size, (int, Sequence)):
            raise TypeError("Size should be int or sequence. Got {}".format(
                type(size)))
        if isinstance(size, Sequence) and len(size) not in (1, 2):
            raise ValueError(
                "If size is a sequence, it should have 1 or 2 values")
        self.size = size
        self.interpolation = interpolation

    def forward(self, img, mask):
        """
        Args:
            img (PIL Image or Tensor): Image to be scaled.

        Returns:
            PIL Image or Tensor: Rescaled image.
        """
        img = TF.resize(img, self.size, self.interpolation)
        mask = TF.resize(mask, self.size, Image.NEAREST)
        return img, mask

    def __repr__(self):
        interpolate_str = _pil_interpolation_to_str[self.interpolation]
        return self.__class__.__name__ + '(size={0}, interpolation={1})'.format(
            self.size, interpolate_str)



from .default import DefaultModelConfig


class ModelConfig(DefaultModelConfig):
    def __init__(self):
        super().__init__()
        self.MODEL_NAME = 'SwinB_AOTL'

        self.MODEL_ENCODER = 'swin_base'
        self.MODEL_ENCODER_PRETRAIN = './pretrain_models/swin_base_patch4_window7_224_22k.pth'  # https://github.com/SwinTransformer/storage/releases/download/v1.0.0/swin_base_patch4_window7_224_22k.pth
        self.MODEL_ALIGN_CORNERS = False
        self.MODEL_ENCODER_DIM = [128, 256, 512, 512]  # 4x, 8x, 16x, 16x
        self.MODEL_LSTT_NUM = 3

        self.TRAIN_LONG_TERM_MEM_GAP = 2

        self.TEST_LONG_TERM_MEM_GAP = 5


import torch
import torch.nn.functional as F
from torch import nn


class GroupNorm1D(nn.Module):
    def __init__(self, indim, groups=8):
        super().__init__()
        self.gn = nn.GroupNorm(groups, indim)

    def forward(self, x):
        return self.gn(x.permute(1, 2, 0)).permute(2, 0, 1)


class GNActDWConv2d(nn.Module):
    def __init__(self, indim, gn_groups=32):
        super().__init__()
        self.gn = nn.GroupNorm(gn_groups, indim)
        self.conv = nn.Conv2d(indim,
                              indim,
                              5,
                              dilation=1,
                              padding=2,
                              groups=indim,
                              bias=False)

    def forward(self, x, size_2d):
        h, w = size_2d
        _, bs, c = x.size()
        x = x.view(h, w, bs, c).permute(2, 3, 0, 1)
        x = self.gn(x)
        x = F.gelu(x)
        x = self.conv(x)
        x = x.view(bs, c, h * w).permute(2, 0, 1)
        return x


class DWConv2d(nn.Module):
    def __init__(self, indim, dropout=0.1):
        super().__init__()
        self.conv = nn.Conv2d(indim,
                              indim,
                              5,
                              dilation=1,
                              padding=2,
                              groups=indim,
                              bias=False)
        self.dropout = nn.Dropout2d(p=dropout, inplace=True)

    def forward(self, x, size_2d):
        h, w = size_2d
        _, bs, c = x.size()
        x = x.view(h, w, bs, c).permute(2, 3, 0, 1)
        x = self.conv(x)
        x = self.dropout(x)
        x = x.view(bs, c, h * w).permute(2, 0, 1)
        return x


class ScaleOffset(nn.Module):
    def __init__(self, indim):
        super().__init__()
        self.gamma = nn.Parameter(torch.ones(indim))
        # torch.nn.init.normal_(self.gamma, std=0.02)
        self.beta = nn.Parameter(torch.zeros(indim))

    def forward(self, x):
        if len(x.size()) == 3:
            return x * self.gamma + self.beta
        else:
            return x * self.gamma.view(1, -1, 1, 1) + self.beta.view(
                1, -1, 1, 1)


class ConvGN(nn.Module):
    def __init__(self, indim, outdim, kernel_size, gn_groups=8):
        super().__init__()
        self.conv = nn.Conv2d(indim,
                              outdim,
                              kernel_size,
                              padding=kernel_size // 2)
        self.gn = nn.GroupNorm(gn_groups, outdim)

    def forward(self, x):
        return self.gn(self.conv(x))


def seq_to_2d(tensor, size_2d):
    h, w = size_2d
    _, n, c = tensor.size()
    tensor = tensor.view(h, w, n, c).permute(2, 3, 0, 1).contiguous()
    return tensor


def drop_path(x, drop_prob: float = 0., training: bool = False):
    if drop_prob == 0. or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (
        x.shape[0],
        x.shape[1],
    ) + (1, ) * (x.ndim - 2
                 )  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(
        shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x.div(keep_prob) * random_tensor
    return output


def mask_out(x, y, mask_rate=0.15, training=False):
    if mask_rate == 0. or not training:
        return x

    keep_prob = 1 - mask_rate
    shape = (
        x.shape[0],
        x.shape[1],
    ) + (1, ) * (x.ndim - 2
                 )  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(
        shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    output = x * random_tensor + y * (1 - random_tensor)

    return output


class DropPath(nn.Module):
    def __init__(self, drop_prob=None, batch_dim=0):
        super(DropPath, self).__init__()
        self.drop_prob = drop_prob
        self.batch_dim = batch_dim

    def forward(self, x):
        return self.drop_path(x, self.drop_prob)

    def drop_path(self, x, drop_prob):
        if drop_prob == 0. or not self.training:
            return x
        keep_prob = 1 - drop_prob
        shape = [1 for _ in range(x.ndim)]
        shape[self.batch_dim] = x.shape[self.batch_dim]
        random_tensor = keep_prob + torch.rand(
            shape, dtype=x.dtype, device=x.device)
        random_tensor.floor_()  # binarize
        output = x.div(keep_prob) * random_tensor
        return output


class DropOutLogit(nn.Module):
    def __init__(self, drop_prob=None):
        super(DropOutLogit, self).__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return self.drop_logit(x, self.drop_prob)

    def drop_logit(self, x, drop_prob):
        if drop_prob == 0. or not self.training:
            return x
        random_tensor = drop_prob + torch.rand(
            x.shape, dtype=x.dtype, device=x.device)
        random_tensor.floor_()  # binarize
        mask = random_tensor * 1e+8 if (
            x.dtype == torch.float32) else random_tensor * 1e+4
        output = x - mask
        return output



from .default_deaot import DefaultModelConfig


class ModelConfig(DefaultModelConfig):
    def __init__(self):
        super().__init__()
        self.MODEL_NAME = 'DeAOTB'

        self.MODEL_LSTT_NUM = 3



from torch import nn
from torch import Tensor
from typing import Callable, Optional, List
from castle.aot.utils.learning import freeze_params

__all__ = ['MobileNetV2']


def _make_divisible(v: float,
                    divisor: int,
                    min_value: Optional[int] = None) -> int:
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class ConvBNActivation(nn.Sequential):
    def __init__(
        self,
        in_planes: int,
        out_planes: int,
        kernel_size: int = 3,
        stride: int = 1,
        groups: int = 1,
        padding: int = -1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        activation_layer: Optional[Callable[..., nn.Module]] = None,
        dilation: int = 1,
    ) -> None:
        if padding == -1:
            padding = (kernel_size - 1) // 2 * dilation
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if activation_layer is None:
            activation_layer = nn.ReLU6
        super().__init__(
            nn.Conv2d(in_planes,
                      out_planes,
                      kernel_size,
                      stride,
                      padding,
                      dilation=dilation,
                      groups=groups,
                      bias=False), norm_layer(out_planes),
            activation_layer(inplace=True))
        self.out_channels = out_planes


# necessary for backwards compatibility
ConvBNReLU = ConvBNActivation


class InvertedResidual(nn.Module):
    def __init__(
            self,
            inp: int,
            oup: int,
            stride: int,
            dilation: int,
            expand_ratio: int,
            norm_layer: Optional[Callable[..., nn.Module]] = None) -> None:
        super(InvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        if norm_layer is None:
            norm_layer = nn.BatchNorm2d

        self.kernel_size = 3
        self.dilation = dilation

        hidden_dim = int(round(inp * expand_ratio))
        self.use_res_connect = self.stride == 1 and inp == oup

        layers: List[nn.Module] = []
        if expand_ratio != 1:
            # pw
            layers.append(
                ConvBNReLU(inp,
                           hidden_dim,
                           kernel_size=1,
                           norm_layer=norm_layer))
        layers.extend([
            # dw
            ConvBNReLU(hidden_dim,
                       hidden_dim,
                       stride=stride,
                       dilation=dilation,
                       groups=hidden_dim,
                       norm_layer=norm_layer),
            # pw-linear
            nn.Conv2d(hidden_dim, oup, 1, 1, 0, bias=False),
            norm_layer(oup),
        ])
        self.conv = nn.Sequential(*layers)
        self.out_channels = oup
        self._is_cn = stride > 1

    def forward(self, x: Tensor) -> Tensor:
        if self.use_res_connect:
            return x + self.conv(x)
        else:
            return self.conv(x)


class MobileNetV2(nn.Module):
    def __init__(self,
                 output_stride=8,
                 norm_layer: Optional[Callable[..., nn.Module]] = None,
                 width_mult: float = 1.0,
                 inverted_residual_setting: Optional[List[List[int]]] = None,
                 round_nearest: int = 8,
                 block: Optional[Callable[..., nn.Module]] = None,
                 freeze_at=0) -> None:
        """
        MobileNet V2 main class
        Args:
            num_classes (int): Number of classes
            width_mult (float): Width multiplier - adjusts number of channels in each layer by this amount
            inverted_residual_setting: Network structure
            round_nearest (int): Round the number of channels in each layer to be a multiple of this number
            Set to 1 to turn off rounding
            block: Module specifying inverted residual building block for mobilenet
            norm_layer: Module specifying the normalization layer to use
        """
        super(MobileNetV2, self).__init__()

        if block is None:
            block = InvertedResidual

        if norm_layer is None:
            norm_layer = nn.BatchNorm2d

        last_channel = 1280
        input_channel = 32
        current_stride = 1
        rate = 1

        if inverted_residual_setting is None:
            inverted_residual_setting = [
                # t, c, n, s
                [1, 16, 1, 1],
                [6, 24, 2, 2],
                [6, 32, 3, 2],
                [6, 64, 4, 2],
                [6, 96, 3, 1],
                [6, 160, 3, 2],
                [6, 320, 1, 1],
            ]

        # only check the first element, assuming user knows t,c,n,s are required
        if len(inverted_residual_setting) == 0 or len(
                inverted_residual_setting[0]) != 4:
            raise ValueError("inverted_residual_setting should be non-empty "
                             "or a 4-element list, got {}".format(
                                 inverted_residual_setting))

        # building first layer
        input_channel = _make_divisible(input_channel * width_mult,
                                        round_nearest)
        self.last_channel = _make_divisible(
            last_channel * max(1.0, width_mult), round_nearest)
        features: List[nn.Module] = [
            ConvBNReLU(3, input_channel, stride=2, norm_layer=norm_layer)
        ]
        current_stride *= 2
        # building inverted residual blocks
        for t, c, n, s in inverted_residual_setting:
            if current_stride == output_stride:
                stride = 1
                dilation = rate
                rate *= s
            else:
                stride = s
                dilation = 1
                current_stride *= s
            output_channel = _make_divisible(c * width_mult, round_nearest)
            for i in range(n):
                if i == 0:
                    features.append(
                        block(input_channel, output_channel, stride, dilation,
                              t, norm_layer))
                else:
                    features.append(
                        block(input_channel, output_channel, 1, rate, t,
                              norm_layer))
                input_channel = output_channel

        # building last several layers
        features.append(
            ConvBNReLU(input_channel,
                       self.last_channel,
                       kernel_size=1,
                       norm_layer=norm_layer))
        # make it nn.Sequential
        self.features = nn.Sequential(*features)

        self._initialize_weights()

        feature_4x = self.features[0:4]
        feautre_8x = self.features[4:7]
        feature_16x = self.features[7:14]
        feature_32x = self.features[14:]

        self.stages = [feature_4x, feautre_8x, feature_16x, feature_32x]

        self.freeze(freeze_at)

    def forward(self, x):
        xs = []
        for stage in self.stages:
            x = stage(x)
            xs.append(x)
        return xs

    def _initialize_weights(self):
        # weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def freeze(self, freeze_at):
        if freeze_at >= 1:
            for m in self.stages[0][0]:
                freeze_params(m)

        for idx, stage in enumerate(self.stages, start=2):
            if freeze_at >= idx:
                freeze_params(stage)



"""
API 調用結構圖
────────────────────────────────────────────────────────────
                           create_extract_ui
                                   │
        ┌──────────────────────────┴──────────────────────────┐
        │                                                     │
  UI "Apply" 按鈕                                        UI "Extract" / "Extract Crop Video" / "Extract Rotation Latent"
        │                                                     │
        ▼                                                     ▼
 setting_preprocess                                  ┌─────────────────────────────────┐
        │                                           │ extract_roi_latent               │
        │                                           │ extract_roi_crop_video           │
        │                                           │ extract_rotation_latent          │
        ▼                                           └─────────────────────────────────┘
  Preprocess (類別)                                          │
        │                                                  ▼
        │                                      ┌─────────────────────────────────┐
        │                                      │ generate_dinov2 (observer)       │
        │                                      └─────────────────────────────────┘
        │                                                  │
        │                                                  ▼
        │                                   ┌─────────────────────────────────┐
        │                                   │  extract_roi_latent_from_video   │
        │                                   │  _extract_roi_crop_video         │
        │                                   │  extract_roi_rotation_latent_from_video  │
        │                                   └─────────────────────────────────┘
        │                                                  │
        │                                                  ▼
        │                                    Observer.extract_batch_latent /
        │                                    Observer.extract_image_latent
        │                                                  │
        └──────────────────────────────────────────────────┘

其他輔助函數：
  - init_select_video_list: 初始化影片選單
  - load_project_config: 統一讀取專案設定與目錄路徑
────────────────────────────────────────────────────────────
"""

import os
import json
import numpy as np
import gradio as gr
from castle import generate_dinov2
from castle.utils.video_io import ReadArray, WriteArray
from castle.utils.h5_io import H5IO
from castle.utils.video_align import (
    center_roi, rotate_based_on_roi_closest_center_point,
    crop, blank_page, rotate_based_on_deg
)
from castle.utils.plot import generate_mix_image

# ---------------------------
# 輔助函數：載入專案設定
# ---------------------------
def load_project_config(storage_path, project_name):
    project_path = os.path.join(storage_path, project_name)
    config_path = os.path.join(project_path, 'config.json')
    with open(config_path, 'r') as f:
        config = json.load(f)
    latent_dir_path = os.path.join(project_path, 'latent')
    os.makedirs(latent_dir_path, exist_ok=True)
    return project_path, config, config_path, latent_dir_path

# ---------------------------
# 初始化影片下拉選單
# ---------------------------
def init_select_video_list(storage_path, project_name):
    if not project_name:
        return gr.update(choices=[])
    project_path, config, _, _ = load_project_config(storage_path, project_name)
    if 'source' not in config:
        return gr.update(choices=[])
    choices = sorted(config['source'])
    choices.append("All")
    return gr.update(choices=choices)

# ---------------------------
# ROI Crop Video 提取
# ---------------------------
def _extract_roi_crop_video(out_path, observer, source_video, tracker, select_roi, preprocess, progress):
    fps = source_video.fps
    crf = 15
    writer = WriteArray(out_path, fps, crf)
    for i in progress.tqdm(range(len(source_video))):
        try:
            frame = source_video[i]
            mask = tracker.read_mask(i)
            processed_frame, _ = preprocess.transform(frame, mask)
            writer.append(processed_frame)
        except Exception as e:
            print(f"Error processing frame {i}: {e}")
            writer.close()
            return False
    writer.close()
    return True

def extract_roi_crop_video(storage_path, project_name, select_model, select_roi, select_video, batch_size, preprocess, progress=gr.Progress()):
    select_roi = int(select_roi)
    project_path, config, _, latent_dir_path = load_project_config(storage_path, project_name)
    observer = generate_dinov2(model_type='dinov2_vitb14_reg')
    config['observer_dim'] = observer.n_feature

    # 決定要處理的影片
    video_list = sorted(config['source']) if select_video == "All" else [select_video]

    out_video_path = ""
    for video_name in video_list:
        source_path = os.path.join(storage_path, project_name, 'sources', video_name)
        source_video = ReadArray(source_path)
        track_dir_path = os.path.join(project_path, 'track', video_name)
        mask_list_path = os.path.join(track_dir_path, 'mask_list.h5')
        tracker = H5IO(mask_list_path)
        base_name = os.path.splitext(source_video.video_name)[0]
        out_video_path = os.path.join(latent_dir_path, f'{base_name}_ROI_{select_roi}_crop.mp4')
        if not _extract_roi_crop_video(out_video_path, observer, source_video, tracker, select_roi, preprocess, progress):
            gr.Info(f"_extract_roi_crop_video fail for video: {video_name}")
    return out_video_path

# ---------------------------
# ROI Latent 提取（一般模式）
# ---------------------------
def extract_roi_latent_from_video(observer, source_video, tracker, batch_size, select_roi, preprocess, progress):
    latent_list = []
    batch_size = int(batch_size)
    total_frames = len(source_video)
    for i in progress.tqdm(range(0, total_frames, batch_size)):
        frames, masks = [], []
        for j in range(batch_size):
            idx = i + j
            if idx >= total_frames:
                break
            frames.append(source_video[idx])
            masks.append(tracker.read_mask(idx))
        processed_frames, processed_masks = [], []
        for frame, mask in zip(frames, masks):
            pf, pm = preprocess.transform(frame, mask)
            processed_frames.append(pf)
            processed_masks.append(pm)
        try:
            latent_batch = observer.extract_batch_latent(processed_frames, processed_masks, select_roi)
            latent_list.extend(latent_batch)
        except Exception as e:
            print(f"Batch starting at frame {i} failed: {e}. Process individually.")
            for idx, (pf, pm) in enumerate(zip(processed_frames, processed_masks)):
                try:
                    latent = observer.extract_image_latent(pf, pm, select_roi)
                except Exception as ex:
                    print(f"Failed at frame {i+idx}: {ex}")
                    latent = observer.nan_latent()
                latent_list.append(latent)
    latent_array = np.array(latent_list)
    print('Extracted latent shape:', latent_array.shape)
    return latent_array

def extract_roi_latent(storage_path, project_name, select_model, select_roi, select_video, batch_size, preprocess, progress=gr.Progress()):
    select_roi = int(select_roi)
    project_path, config, config_path, latent_dir_path = load_project_config(storage_path, project_name)
    observer = generate_dinov2(model_type='dinov2_vitb14_reg')
    config['observer_dim'] = observer.n_feature

    video_list = sorted(config['source']) if select_video == "All" else [select_video]
    latent_file_list = []
    for video_name in video_list:
        source_path = os.path.join(storage_path, project_name, 'sources', video_name)
        source_video = ReadArray(source_path)
        track_dir_path = os.path.join(project_path, 'track', video_name)
        mask_list_path = os.path.join(track_dir_path, 'mask_list.h5')
        tracker = H5IO(mask_list_path)
        latent = extract_roi_latent_from_video(observer, source_video, tracker, batch_size, select_roi, preprocess, progress)
        base_name = os.path.splitext(source_video.video_name)[0]
        latent_path = os.path.join(latent_dir_path, f'{base_name}_ROI_{select_roi}_latent.npz')
        np.savez_compressed(latent_path, latent=latent)
        config.setdefault('latent', {})[f'{base_name}_ROI_{select_roi}_latent.npz'] = source_video.video_name
        with open(config_path, 'w') as f:
            json.dump(config, f)
        latent_file_list.append(latent_path)
    return latent_file_list

# ---------------------------
# ROI Rotation Latent 提取
# ---------------------------
def extract_roi_rotation_latent_from_video(observer, source_video, tracker, batch_size, select_roi, preprocess, progress):
    latent_list = []
    batch_size = int(batch_size)
    total_frames = len(source_video)
    for i in progress.tqdm(range(0, total_frames, batch_size)):
        frames, masks = [], []
        for j in range(batch_size):
            idx = i + j
            
            if idx >= total_frames:
                break
            frame = source_video[idx]
            mask = tracker.read_mask(idx)
            # 對每個 frame 依據不同旋轉角度進行處理
            for deg in range(0, 360, 15):
                pf, pm = preprocess.transform(frame, mask, deg)
                frames.append(pf)
                masks.append(pm)
        latent_batch = observer.extract_batch_latent(frames, masks, select_roi)
        latent_batch = np.array(latent_batch)
        latent_batch = latent_batch.reshape(len(latent_batch) // 24, 24, 768).mean(axis=1)
        latent_list.extend(latent_batch)
    latent_array = np.array(latent_list)
    print('Extracted rotation latent shape:', latent_array.shape)
    return latent_array

def extract_rotation_latent(storage_path, project_name, select_model, select_roi, select_video, batch_size, preprocess, progress=gr.Progress()):
    select_roi = int(select_roi)
    project_path, config, config_path, latent_dir_path = load_project_config(storage_path, project_name)
    observer = generate_dinov2(model_type='dinov2_vitb14_reg')
    config['observer_dim'] = observer.n_feature

    video_list = sorted(config['source']) if select_video == "All" else [select_video]
    latent_file_list = []
    for video_name in video_list:
        source_path = os.path.join(storage_path, project_name, 'sources', video_name)
        source_video = ReadArray(source_path)
        track_dir_path = os.path.join(project_path, 'track', video_name)
        mask_list_path = os.path.join(track_dir_path, 'mask_list.h5')
        tracker = H5IO(mask_list_path)
        latent = extract_roi_rotation_latent_from_video(observer, source_video, tracker, batch_size, select_roi, preprocess, progress)
        base_name = os.path.splitext(source_video.video_name)[0]
        latent_path = os.path.join(latent_dir_path, f'{base_name}_ROI_{select_roi}_rotation_latent.npz')
        np.savez_compressed(latent_path, latent=latent)
        config.setdefault('latent', {})[f'{base_name}_ROI_{select_roi}_rotation_latent.npz'] = source_video.video_name
        with open(config_path, 'w') as f:
            json.dump(config, f)
        latent_file_list.append(latent_path)
    return latent_file_list

# ---------------------------
# 預處理類別
# ---------------------------
class Preprocess:
    def __init__(self, center_roi_switch, center_roi_id, center_roi_crop_width, center_roi_crop_height, rotate_roi_tail_switch, rotate_roi_tail_id, remove_background_switch='False'):
        self.center_roi_switch = (center_roi_switch == 'True')
        self.center_roi_id = center_roi_id
        self.center_roi_crop_width = int(center_roi_crop_width)
        self.center_roi_crop_height = int(center_roi_crop_height)
        self.rotate_roi_tail_switch = (rotate_roi_tail_switch == 'True')
        self.rotate_roi_tail_id = rotate_roi_tail_id
        self.remove_background_switch = (remove_background_switch == 'True')

    def transform(self, frame, mask, deg=0):
        try:
            if self.center_roi_switch:
                f = center_roi(frame, mask, self.center_roi_id)
                m = center_roi(mask, mask, self.center_roi_id)
                if self.rotate_roi_tail_switch:
                    f = rotate_based_on_roi_closest_center_point(f, m, self.rotate_roi_tail_id)
                    m = rotate_based_on_roi_closest_center_point(m, m, self.rotate_roi_tail_id)
                if deg > 0:
                    f = rotate_based_on_deg(f, deg)
                    m = rotate_based_on_deg(m, deg)
                f = crop(f, self.center_roi_crop_height, self.center_roi_crop_width)
                m = crop(m, self.center_roi_crop_height, self.center_roi_crop_width)
            else:
                f, m = frame, mask
            if self.remove_background_switch:
                f[m == 0] = 255
        except Exception as e:
            print(f"Error in Preprocess.transform: {e}")
            f = blank_page(self.center_roi_crop_height, self.center_roi_crop_width)
            m = blank_page(self.center_roi_crop_height, self.center_roi_crop_width)
        return f, m

def setting_preprocess(storage_path, project_name, select_video, center_roi_switch, center_roi_id,
                       center_roi_crop_width, center_roi_crop_height, rotate_roi_tail_switch, rotate_roi_tail_id, remove_background_switch):
    preprocess = Preprocess(center_roi_switch, center_roi_id, center_roi_crop_width,
                            center_roi_crop_height, rotate_roi_tail_switch, rotate_roi_tail_id, remove_background_switch)
    project_path, config, _, _ = load_project_config(storage_path, project_name)
    video_list = sorted(config['source']) if select_video == "All" else [select_video]
    first_video = video_list[0]
    source_video = ReadArray(os.path.join(storage_path, project_name, 'sources', first_video))
    track_dir_path = os.path.join(project_path, 'track', first_video)
    mask_list_path = os.path.join(track_dir_path, 'mask_list.h5')
    tracker = H5IO(mask_list_path)
    frame, mask = source_video[0], tracker.read_mask(0)
    processed_frame, processed_mask = preprocess.transform(frame, mask)
    mixed_image = generate_mix_image(processed_frame, processed_mask)
    del tracker
    return preprocess, mixed_image

# ---------------------------
# 建立 Gradio UI
# ---------------------------
def create_extract_ui(storage_path, project_name, extract_tab):
    ui = {}
    preprocess_state = gr.State(None)
    with gr.Row(visible=True):
        with gr.Column(scale=2):
            ui['select_model'] = gr.Dropdown(
                label="Select Visual Model",
                choices=["dinov2_vitb14_reg4_pretrain"],
                value="dinov2_vitb14_reg4_pretrain",
                visible=False
            )
            ui['select_roi_id'] = gr.Textbox(
                label="Enter ROI ID", value="1", info="ex: 1,2,3.", visible=False
            )
            ui['batch_size'] = gr.Textbox(
                label="Batch size", value="8", info="ex: 1, 16, 64 ...", visible=False
            )
            ui['select_video'] = gr.Dropdown(
                label="Select Target Video", value='All', visible=False
            )
        with gr.Column(scale=2):
            ui['center_roi_switch'] = gr.Dropdown(
                label="Center ROI", value='False', choices=['True', 'False'], visible=False
            )
            ui['center_roi_id'] = gr.Number(label="Center ROI ID", value=1, visible=False)
            ui['center_roi_crop_width'] = gr.Number(label="width", value=300, visible=False)
            ui['center_roi_crop_height'] = gr.Number(label="height", value=300, visible=False)
            ui['rotate_roi_tail_switch'] = gr.Dropdown(
                label="Rotate based on Tail", value='False', choices=['True', 'False'], visible=False
            )
            ui['rotate_roi_tail_id'] = gr.Number(label="Tail ROI ID", value=2, visible=False)
            ui['remove_background_switch'] = gr.Dropdown(
                label="Remove Background", value='False', choices=['True', 'False'], visible=False
            )
            ui['apply_preprocess'] = gr.Button("Apply", visible=False)
        with gr.Column(scale=4):
            ui['display'] = gr.Image(label='Display', interactive=False, visible=False)
            ui['extract_btn'] = gr.Button("Extract", visible=False)
            ui['extract_crop_video_btn'] = gr.Button("Extract Crop Video", visible=False)
            ui['extract_rotation_latent_btn'] = gr.Button("Extract Rotation Latent", visible=False)
            ui['latent_file_list'] = gr.File(label="ROI Visual Representation File List", visible=False)

    extract_tab.select(
        fn=init_select_video_list,
        inputs=[storage_path, project_name],
        outputs=ui['select_video']
    )
    ui['extract_crop_video_btn'].click(
        fn=extract_roi_crop_video,
        inputs=[storage_path, project_name, ui['select_model'], ui['select_roi_id'],
                ui['select_video'], ui['batch_size'], preprocess_state],
        outputs=ui['latent_file_list']
    )
    ui['extract_btn'].click(
        fn=extract_roi_latent,
        inputs=[storage_path, project_name, ui['select_model'], ui['select_roi_id'],
                ui['select_video'], ui['batch_size'], preprocess_state],
        outputs=ui['latent_file_list']
    )
    ui['extract_rotation_latent_btn'].click(
        fn=extract_rotation_latent,
        inputs=[storage_path, project_name, ui['select_model'], ui['select_roi_id'],
                ui['select_video'], ui['batch_size'], preprocess_state],
        outputs=ui['latent_file_list']
    )
    ui['apply_preprocess'].click(
        fn=setting_preprocess,
        inputs=[storage_path, project_name, ui['select_video'], ui['center_roi_switch'],
                ui['center_roi_id'], ui['center_roi_crop_width'], ui['center_roi_crop_height'],
                ui['rotate_roi_tail_switch'], ui['rotate_roi_tail_id'], ui['remove_background_switch']],
        outputs=[preprocess_state, ui['display']]
    )
    return ui



import os
import gradio as gr
import shutil
import json


def upload_local_videos(storage_path, project_name, upload_video_path):
    project_path = os.path.join(storage_path, project_name)
    project_config_path = os.path.join(project_path, 'config.json')
    source_dir_path = os.path.join(project_path, 'sources/')
    os.makedirs(source_dir_path, exist_ok=True)
    project_config = json.load(open(project_config_path, 'r'))
    if not 'source' in project_config:
            project_config['source'] = []

    for it in upload_video_path:
        if os.path.basename(it.name) in project_config['source']:
             gr.Info("Already in this project")
             continue
        shutil.copyfile(it.name, os.path.join(source_dir_path, os.path.basename(it.name)))
        project_config['source'].append(os.path.basename(it.name))
    json.dump(project_config, open(project_config_path,'w'))

def update_project_video_info(storage_path, project_name):
    project_path = os.path.join(storage_path, project_name)
    project_config_path = os.path.join(project_path, 'config.json')
    project_config = json.load(open(project_config_path, 'r'))
    if not 'source' in project_config:
        return
    return
     
#          return ""
#     return str(project_config['source'])


video_extensions = ['.mp4', '.avi', '.mov', '.wmv', '.flv', '.mkv']

# Function to check if a file is a video
def is_video(file_name):
    _, ext = os.path.splitext(file_name)
    return ext.lower() in video_extensions

def list_server_video(storage_path):
    subdirectories = sorted([d for d in os.listdir(storage_path) if is_video(os.path.join(storage_path, d))])
    return gr.update(choices=subdirectories)

def upload_server_video(storage_path, project_name, video_storage_path, select_server_video):
     project_path = os.path.join(storage_path, project_name)
     project_config_path = os.path.join(project_path, 'config.json')
     source_dir_path = os.path.join(project_path, 'sources/')
     os.makedirs(source_dir_path, exist_ok=True)
     project_config = json.load(open(project_config_path, 'r'))
     if not 'source' in project_config:
          project_config['source'] = []

     if os.path.basename(select_server_video) in project_config['source']:
          gr.Info("Already in this project")
          return
     
     shutil.copyfile(os.path.join(video_storage_path, select_server_video), os.path.join(source_dir_path, select_server_video))
     project_config['source'].append(os.path.basename(select_server_video))
     json.dump(project_config, open(project_config_path,'w'))


def create_source_ui(storage_path, project_name):
    ui = dict()
    
#     ui['project_video_info'] = gr.Textbox(
#          label='Project Video Information', 
#          interactive=False,visible=False)
#     gr.HTML(value="<br>")
    ui['upload_local_videos'] = gr.File(label="Add Local Videos", file_types=["video"], interactive=True, file_count="multiple", visible=False)
    gr.HTML(value="<br>")
    # ui['upload_another'] = gr.Button('Upload Another Video', interactive=True, visible=False)
    ui['storage_path'] = gr.Textbox(
         label='Add Server Videos: Storage Location',
         info='The location which storage videos',
         value='demo/',
         interactive=True, visible=False)
    ui['select_server_video'] = gr.Dropdown(
         label="Add Server Video", 
         interactive=True, visible=False)
    
    ui['upload_local_videos'].upload(
         fn=upload_local_videos,
         inputs=[
              storage_path, project_name, 
              ui['upload_local_videos']]
    )

    ui['upload_local_videos'].upload(
         fn=update_project_video_info,
         inputs=[storage_path, project_name],
     #     outputs=ui['project_video_info']
    )

    ui['select_server_video'].focus(
         fn=list_server_video,
         inputs=ui['storage_path'],
         outputs=ui['select_server_video']
    )
    ui['select_server_video'].select(
        fn=upload_server_video,
        inputs=[storage_path, project_name, 
              ui['storage_path'], ui['select_server_video']]
    )

    return ui


import torch.nn as nn

from castle.aot.networks.encoders import build_encoder
from castle.aot.networks.layers.transformer import LongShortTermTransformer
from castle.aot.networks.decoders import build_decoder
from castle.aot.networks.layers.position import PositionEmbeddingSine


class AOT(nn.Module):
    def __init__(self, cfg, encoder='mobilenetv2', decoder='fpn'):
        super().__init__()
        self.cfg = cfg
        self.max_obj_num = cfg.MODEL_MAX_OBJ_NUM
        self.epsilon = cfg.MODEL_EPSILON

        self.encoder = build_encoder(encoder,
                                     frozen_bn=cfg.MODEL_FREEZE_BN,
                                     freeze_at=cfg.TRAIN_ENCODER_FREEZE_AT)
        self.encoder_projector = nn.Conv2d(cfg.MODEL_ENCODER_DIM[-1],
                                           cfg.MODEL_ENCODER_EMBEDDING_DIM,
                                           kernel_size=1)

        self.LSTT = LongShortTermTransformer(
            cfg.MODEL_LSTT_NUM,
            cfg.MODEL_ENCODER_EMBEDDING_DIM,
            cfg.MODEL_SELF_HEADS,
            cfg.MODEL_ATT_HEADS,
            emb_dropout=cfg.TRAIN_LSTT_EMB_DROPOUT,
            droppath=cfg.TRAIN_LSTT_DROPPATH,
            lt_dropout=cfg.TRAIN_LSTT_LT_DROPOUT,
            st_dropout=cfg.TRAIN_LSTT_ST_DROPOUT,
            droppath_lst=cfg.TRAIN_LSTT_DROPPATH_LST,
            droppath_scaling=cfg.TRAIN_LSTT_DROPPATH_SCALING,
            intermediate_norm=cfg.MODEL_DECODER_INTERMEDIATE_LSTT,
            return_intermediate=True)

        decoder_indim = cfg.MODEL_ENCODER_EMBEDDING_DIM * \
            (cfg.MODEL_LSTT_NUM +
             1) if cfg.MODEL_DECODER_INTERMEDIATE_LSTT else cfg.MODEL_ENCODER_EMBEDDING_DIM

        self.decoder = build_decoder(
            decoder,
            in_dim=decoder_indim,
            out_dim=cfg.MODEL_MAX_OBJ_NUM + 1,
            decode_intermediate_input=cfg.MODEL_DECODER_INTERMEDIATE_LSTT,
            hidden_dim=cfg.MODEL_ENCODER_EMBEDDING_DIM,
            shortcut_dims=cfg.MODEL_ENCODER_DIM,
            align_corners=cfg.MODEL_ALIGN_CORNERS)

        if cfg.MODEL_ALIGN_CORNERS:
            self.patch_wise_id_bank = nn.Conv2d(
                cfg.MODEL_MAX_OBJ_NUM + 1,
                cfg.MODEL_ENCODER_EMBEDDING_DIM,
                kernel_size=17,
                stride=16,
                padding=8)
        else:
            self.patch_wise_id_bank = nn.Conv2d(
                cfg.MODEL_MAX_OBJ_NUM + 1,
                cfg.MODEL_ENCODER_EMBEDDING_DIM,
                kernel_size=16,
                stride=16,
                padding=0)

        self.id_dropout = nn.Dropout(cfg.TRAIN_LSTT_ID_DROPOUT, True)

        self.pos_generator = PositionEmbeddingSine(
            cfg.MODEL_ENCODER_EMBEDDING_DIM // 2, normalize=True)

        self._init_weight()

    def get_pos_emb(self, x):
        pos_emb = self.pos_generator(x)
        return pos_emb

    def get_id_emb(self, x):
        id_emb = self.patch_wise_id_bank(x)
        id_emb = self.id_dropout(id_emb)
        return id_emb

    def encode_image(self, img):
        xs = self.encoder(img)
        xs[-1] = self.encoder_projector(xs[-1])
        return xs

    def decode_id_logits(self, lstt_emb, shortcuts):
        n, c, h, w = shortcuts[-1].size()
        decoder_inputs = [shortcuts[-1]]
        for emb in lstt_emb:
            decoder_inputs.append(emb.view(h, w, n, c).permute(2, 3, 0, 1))
        pred_logit = self.decoder(decoder_inputs, shortcuts)
        return pred_logit

    def LSTT_forward(self,
                     curr_embs,
                     long_term_memories,
                     short_term_memories,
                     curr_id_emb=None,
                     pos_emb=None,
                     size_2d=(30, 30)):
        n, c, h, w = curr_embs[-1].size()
        curr_emb = curr_embs[-1].view(n, c, h * w).permute(2, 0, 1)
        lstt_embs, lstt_memories = self.LSTT(curr_emb, long_term_memories,
                                             short_term_memories, curr_id_emb,
                                             pos_emb, size_2d)
        lstt_curr_memories, lstt_long_memories, lstt_short_memories = zip(
            *lstt_memories)
        return lstt_embs, lstt_curr_memories, lstt_long_memories, lstt_short_memories

    def _init_weight(self):
        nn.init.xavier_uniform_(self.encoder_projector.weight)
        nn.init.orthogonal_(
            self.patch_wise_id_bank.weight.view(
                self.cfg.MODEL_ENCODER_EMBEDDING_DIM, -1).permute(0, 1),
            gain=17**-2 if self.cfg.MODEL_ALIGN_CORNERS else 16**-2)



import os
import cv2
import numpy as np

import gradio as gr

from .plot_mask_info import Plotter
from castle.utils.plot import generate_mix_image, generate_mask_image
from castle.utils.h5_io import H5IO
from castle.utils.video_io import ReadArray, WriteArray

def plot_basic_mask_info(storage_path, project_name, source_video, progress=gr.Progress()):
    project_path = os.path.join(storage_path, project_name)
    video_name = source_video.video_name
    track_dir_path = os.path.join(project_path, 'track', video_name)

    rois_results_path = os.path.join(track_dir_path, f'mask_list.h5')
    rois_results = H5IO(rois_results_path) # TODO: check file is exist first
    n_rois = rois_results.get_n_rois()

    total_frames = len(rois_results)
    roi_info_list = [{"x":[], "y":[], "area":[]} for i in range(n_rois)]
    for i_frame in progress.tqdm(range(total_frames)):
        for i in range(n_rois):
            if not rois_results.has_mask(i_frame):
                roi_info_list[i]['x'].append(np.nan)
                roi_info_list[i]['y'].append(np.nan)
                roi_info_list[i]['area'].append(0)
                continue


            mask = rois_results[i_frame][:]
            mask = cv2.inRange(mask, i+1, i+1)
            output = cv2.connectedComponentsWithStats(mask, 8, cv2.CV_32S)
            num_labels, _, stats, centroids = output
            if num_labels <= 1:
                roi_info_list[i]['x'].append(np.nan)
                roi_info_list[i]['y'].append(np.nan)
                roi_info_list[i]['area'].append(0)
                continue


            areas = [stats[j, cv2.CC_STAT_AREA] for j in range(1, num_labels)]
            max_label = np.argmax(areas)
            roi_info_list[i]['x'].append(round(centroids[max_label + 1][0]))
            roi_info_list[i]['y'].append(round(centroids[max_label + 1][1]))
            roi_info_list[i]['area'].append(areas[max_label])
            
    for i in range(n_rois):
        roi_info_list[i]['x'] = np.array(roi_info_list[i]['x'])
        roi_info_list[i]['y'] = np.array(roi_info_list[i]['y'])
        roi_info_list[i]['area'] = np.array(roi_info_list[i]['area']).astype(int)

    del rois_results
    return Plotter.plot_position(roi_info_list), Plotter.plot_speed(roi_info_list), Plotter.plot_area(roi_info_list), roi_info_list


def generate_mask_kinematic_csv(storage_path, project_name, source_video, roi_info_list):
    project_path = os.path.join(storage_path, project_name)
    video_name = source_video.video_name
    track_dir_path = os.path.join(project_path, 'track', video_name)
    video_name_wo_extension = video_name.split('.')[0]
    mask_kinematic_csv_path = os.path.join(track_dir_path, f'{video_name_wo_extension}-basic-infomation.csv')
    df = Plotter.create_pandas(roi_info_list)
    df.to_csv(mask_kinematic_csv_path)
    return mask_kinematic_csv_path


def generate_mask_video(storage_path, project_name, source_video):
    project_path = os.path.join(storage_path, project_name)
    video_name = source_video.video_name
    track_dir_path = os.path.join(project_path, 'track', video_name)
    rois_results_path = os.path.join(track_dir_path, f'mask_list.h5')
    video_name_wo_extension = video_name.split('.')[0]
    output_path = os.path.join(track_dir_path, f'{video_name_wo_extension}-rois.mp4')
    output = WriteArray(output_path, fps=source_video.fps, crf=15)
    rois_results = H5IO(rois_results_path) # TODO: check file is exist first
    n_frames = len(rois_results)

    for i in range(n_frames):
        rois = rois_results[i]
        out_frame = generate_mask_image(rois)
        output.append(out_frame)

    del rois_results, output
    return output_path


def generate_mix_video(storage_path, project_name, source_video):
    project_path = os.path.join(storage_path, project_name)
    video_name = source_video.video_name
    track_dir_path = os.path.join(project_path, 'track', video_name)
    rois_results_path = os.path.join(track_dir_path, f'mask_list.h5')
    video_name_wo_extension = video_name.split('.')[0]
    output_path = os.path.join(track_dir_path, f'{video_name_wo_extension}-mix.mp4')
    output = WriteArray(output_path, fps=source_video.fps, crf=15)
    rois_results = H5IO(rois_results_path) # TODO: check file is exist first
    n_frames = len(rois_results)

    for i in range(n_frames):
        rois = rois_results[i]
        frame = source_video[i]
        out_frame = generate_mix_image(frame, rois)
        output.append(out_frame)

    del rois_results, output
    return output_path


def create_post_track_ui(storage_path, project_name, source_video):
    ui = dict()
    roi_info_list = gr.State(None)
    # with gr.Accordion('Basic Kinematic Infomation', open=True, visible=False) as ui['basic_mask_info_accordion']:
    ui['analysis_mask'] = gr.Button("Analysis Mask", interactive=True, visible=False)
    with gr.Row(visible=True):
        ui['position_plot'] = gr.Plot(label="Position", visible=False)
    with gr.Row(visible=True):
            ui['velocity_plot'] = gr.Plot(label="Speed", visible=False)
    with gr.Row(visible=True):
            ui['area_plot'] = gr.Plot(label="Area", visible=False)  
    with gr.Row(visible=True):
        with gr.Column(scale=2):
            ui['generate_mask_kinematic_btn'] = gr.Button("Generate Basic Kinematic CSV", interactive=False, visible=False)
        with gr.Column(scale=8):
            ui['mask_kinematic_file'] = gr.File(label="Basic Kinematic CSV", interactive=False, visible=False)
    with gr.Row(visible=True):
        with gr.Column(scale=2):
            ui['generate_mask_video_btn'] = gr.Button("Generate ROIs Video", interactive=True, visible=False)
        with gr.Column(scale=8):
            ui['mask_video'] = gr.File(label="ROIs Video", interactive=False, visible=False)
    with gr.Row(visible=True):
        with gr.Column(scale=2):
            ui['generate_mix_video_btn'] = gr.Button("Generate Mix Video", interactive=True, visible=False)
        with gr.Column(scale=8):
            ui['mix_video'] = gr.File(label="Mix Video", interactive=False, visible=False)

    ui['analysis_mask'].click(
        fn=plot_basic_mask_info,
        inputs=[storage_path, project_name, source_video],
        outputs=[ui['position_plot'], ui['velocity_plot'], ui['area_plot'], roi_info_list]
    )

    ui['generate_mask_kinematic_btn'].click(
        fn=generate_mask_kinematic_csv,
        inputs=[storage_path, project_name, source_video, roi_info_list],
        outputs=ui['mask_kinematic_file']
    )

    ui['generate_mask_video_btn'].click(
        fn=generate_mask_video,
        inputs=[storage_path, project_name, source_video],
        outputs=ui['mask_video']
    )

    ui['generate_mix_video_btn'].click(
        fn=generate_mix_video,
        inputs=[storage_path, project_name, source_video],
        outputs=ui['mix_video']
    )
      
    return ui


from castle.aot.networks.models.aot import AOT
from castle.aot.networks.models.deaot import DeAOT


def build_vos_model(name, cfg, **kwargs):
    if name == 'aot':
        return AOT(cfg, encoder=cfg.MODEL_ENCODER, **kwargs)
    elif name == 'deaot':
        return DeAOT(cfg, encoder=cfg.MODEL_ENCODER, **kwargs)
    else:
        raise NotImplementedError



import torch
import torch.nn as nn
import torch.nn.functional as F


class FrozenBatchNorm2d(nn.Module):
    """
    BatchNorm2d where the batch statistics and the affine parameters
    are fixed
    """
    def __init__(self, n, epsilon=1e-5):
        super(FrozenBatchNorm2d, self).__init__()
        self.register_buffer("weight", torch.ones(n))
        self.register_buffer("bias", torch.zeros(n))
        self.register_buffer("running_mean", torch.zeros(n))
        self.register_buffer("running_var", torch.ones(n) - epsilon)
        self.epsilon = epsilon

    def forward(self, x):
        """
        Refer to Detectron2 (https://github.com/facebookresearch/detectron2/blob/cbbc1ce26473cb2a5cc8f58e8ada9ae14cb41052/detectron2/layers/batch_norm.py)
        """
        if x.requires_grad:
            # When gradients are needed, F.batch_norm will use extra memory
            # because its backward op computes gradients for weight/bias as well.
            scale = self.weight * (self.running_var + self.epsilon).rsqrt()
            bias = self.bias - self.running_mean * scale
            scale = scale.reshape(1, -1, 1, 1)
            bias = bias.reshape(1, -1, 1, 1)
            out_dtype = x.dtype  # may be half
            return x * scale.to(out_dtype) + bias.to(out_dtype)
        else:
            # When gradients are not needed, F.batch_norm is a single fused op
            # and provide more optimization opportunities.
            return F.batch_norm(
                x,
                self.running_mean,
                self.running_var,
                self.weight,
                self.bias,
                training=False,
                eps=self.epsilon,
            )



import time
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Generator

import numpy as np
import gradio as gr
from natsort import natsorted

from castle import generate_aot
from castle.utils.h5_io import H5IO
from castle.utils.plot import generate_mix_image


def read_label(
    storage_path: str, project_name: str, source_video: Optional[Any]
) -> List[Dict[str, Any]]:
    """
    Read all label files for the given project and return a list of labels.

    Each label is a dict with keys:
        - index: a string identifier combining file index and video basename.
        - frame: the frame data.
        - mask: the corresponding mask.

    Args:
        storage_path: Base storage directory.
        project_name: Name of the project.
        source_video: Video source object; if None, an empty list is returned.

    Returns:
        A list of dictionaries containing label information.
    """
    if source_video is None:
        return []

    project_path = Path(storage_path) / project_name
    label_dir = project_path / "label"

    label_list = []
    # Iterate through all subdirectories in natural sorted order
    for label_folder in natsorted([p for p in label_dir.iterdir() if p.is_dir()]):
        video_basename = label_folder.name
        # Iterate through all .npz files in the folder
        for npz_file in natsorted(list(label_folder.glob("*.npz"))):
            index = npz_file.stem
            data = np.load(npz_file)
            # Expect keys 'frame' and 'mask'
            if "frame" not in data or "mask" not in data:
                continue
            frame = data["frame"]
            mask = data["mask"]
            label_list.append(
                {
                    "index": f"{index}, {video_basename}",
                    "frame": frame,
                    "mask": mask,
                }
            )
    return label_list


def read_label_to_gallery(
    storage_path: str, project_name: str, source_video: Optional[Any]
) -> Tuple[List[Dict[str, Any]]]:
    """
    Generate a gallery list based on the label data.

    Each gallery entry is a tuple (mixed_image, label_index).

    Args:
        storage_path: Base storage directory.
        project_name: Name of the project.
        source_video: Video source object.

    Returns:
        A tuple containing the original label list and a gallery list.
    """
    label_list = read_label(storage_path, project_name, source_video)
    gallery_list = [
        (generate_mix_image(label["frame"], label["mask"]), label["index"])
        for label in label_list
    ]
    return label_list


class InferenceTracker:
    """
    This class sets up a tracker for performing ROI tracking on a video,
    using previously known frames and masks as references.
    """

    def __init__(
        self,
        storage_path: str,
        project_name: str,
        source_video: Any,
        start: int,
        stop: int,
        model_aot: str,
    ) -> None:
        self.cancel: bool = False
        self.show_middle_result: bool = False
        self.model_aot: str = model_aot

        project_path = Path(storage_path) / project_name
        video_name = source_video.video_name
        self.track_dir_path: Path = project_path / "track" / video_name
        self.track_dir_path.mkdir(parents=True, exist_ok=True)

        self.source_video = source_video
        self.start = int(start)
        self.stop = int(stop)
        self.max_len = 30

        # Prepare reference knowledge from labels
        self.knowledges: List[Tuple[Any, Any]] = []
        label_list = read_label(storage_path, project_name, source_video)
        self.n_rois: int = 0
        for label in label_list:
            frame, mask = label["frame"], label["mask"]
            self.knowledges.append((frame, mask))
            # Update n_rois to be the maximum value found in masks
            self.n_rois = max(self.n_rois, int(np.max(mask)))

    def tracking(self, progress: gr.Progress) -> str:
        """
        Track ROIs over the specified frames and write each frame's result into an H5 file.

        Args:
            progress: A gradio Progress object for displaying progress.

        Returns:
            A status message ("Done" or "Cancel").
        """
        time.sleep(0.5)
        tracker = generate_aot(model_type=self.model_aot)
        mask_list_path = self.track_dir_path / "mask_list.h5"
        mask_seq = H5IO(str(mask_list_path))

        # Write video and ROI configuration settings
        mask_seq.write_config("n_rois", self.n_rois)
        mask_seq.write_config("total_frames", len(self.source_video))
        mask_seq.write_config("height", self.source_video.video_stream.height)
        mask_seq.write_config("width", self.source_video.video_stream.width)
        # Add all reference ROI frames to the tracker’s memory
        for frame, mask in self.knowledges:
            tracker.add_reference_frame(frame, mask, self.n_rois, -1)

        delta = 1 if self.start < self.stop else -1
        for i in progress.tqdm(range(self.start, self.stop + delta, delta)):
            if self.cancel:
                self.show_middle_result = False
                self.cancel = False
                del mask_seq
                return "Cancel"

            frame = self.source_video[i]
            mask = tracker.track(frame)
            tracker.update_memory(mask)
            # Save latest frame and processed mask for potential display
            self.frame = frame
            self.mask = mask.squeeze().detach().cpu().numpy().astype(np.uint8)
            mask_seq.write_mask(i, self.mask)

        self.show_middle_result = False
        del mask_seq
        return "Done"

    def set_cancel(self) -> None:
        """Set the flag to cancel tracking."""
        self.cancel = True

    def flip_show_middle_result(self) -> None:
        """Toggle the display of intermediate results."""
        self.show_middle_result = not self.show_middle_result


def init_inference_tracker(
    storage_path: str,
    project_name: str,
    source_video: Any,
    start: int,
    stop: int,
    model_aot: str,
) -> InferenceTracker:
    """
    Initialize and return an instance of InferenceTracker.
    """
    print("Initializing InferenceTracker with start:", start, "and stop:", stop)
    return InferenceTracker(storage_path, project_name, source_video, start, stop, model_aot)


def run_inference_tracker(
    tracker: InferenceTracker, progress: gr.Progress = gr.Progress()
) -> str:
    """
    Run the tracking process.

    Args:
        tracker: The InferenceTracker instance.
        progress: A gradio Progress instance.

    Returns:
        A status message.
    """
    status = tracker.tracking(progress)
    return f"{status}. From {tracker.start} to {tracker.stop}"


def click_middle_result(
    tracker: InferenceTracker,
) -> Generator[Tuple[Any, str], None, None]:
    """
    Toggle the display of intermediate results and yield updates.

    This generator repeatedly yields a tuple (mixed_image, display_mode) every second.
    """
    print("Toggling middle result display. Current state:", tracker.show_middle_result)
    tracker.flip_show_middle_result()
    while tracker.show_middle_result:
        time.sleep(1)
        mixed_image = generate_mix_image(tracker.frame, tracker.mask)
        yield mixed_image, display_middle_result_mode(tracker.show_middle_result)


def display_middle_result_mode(is_showing: bool) -> str:
    """
    Return the display mode string based on a boolean flag.

    Args:
        is_showing: True if the intermediate result is being shown.

    Returns:
        "Show" if is_showing is True, otherwise "Close".
    """
    return "Show" if is_showing else "Close"


def set_cancel(tracker: InferenceTracker) -> None:
    """Trigger cancellation of the tracking process."""
    tracker.set_cancel()


def create_track_ui(
    storage_path: str, project_name: str, source_video: Any, track_tab: gr.Tab
) -> Dict[str, Any]:
    """
    Create and return the Gradio UI components for tracking.

    Args:
        storage_path: Base storage directory.
        project_name: Name of the project.
        source_video: Video source object.
        track_tab: The Gradio Tab component where UI elements are added.

    Returns:
        A dictionary of UI elements.
    """
    ui: Dict[str, Any] = {}

    label_list_state = gr.State(None)
    tracker_state = gr.State(None)



    with gr.Accordion("Inference", open=True, visible=False) as inference_accordion:
        with gr.Row(visible=True):
            with gr.Column(scale=2):
                start_frame = gr.Slider(
                    label="Start Frame (include)",
                    minimum=0,
                    step=1,
                    maximum=1,
                    value=0,
                    interactive=True,
                    visible=False,
                )
                stop_frame = gr.Slider(
                    label="Stop Frame (include)",
                    minimum=0,
                    step=1,
                    maximum=1,
                    value=1,
                    interactive=True,
                    visible=False,
                )
                model_dropdown = gr.Dropdown(
                    choices=["r50_deaotl", "swinb_deaotl"],
                    label="Tracking Model",
                    info="ResNet-50 or Swin-transformer",
                    value="r50_deaotl",
                    interactive=True,
                )
                init_tracker_btn = gr.Button("Init Tracker", interactive=True, visible=False)
                tracking_btn = gr.Button("Tracking ROIs", interactive=True, visible=False)
                progress_text = gr.Textbox(label="Progress", visible=False)
                display_mode_text = gr.Textbox(
                    value="Close", label="Display Mode", interactive=False, visible=False
                )
                display_middle_result_btn = gr.Button(
                    "Display middle result", interactive=True, visible=False
                )
                cancel_btn = gr.Button("Cancel", interactive=True, visible=False)

            with gr.Column(scale=8):
                display = gr.Image(label="Display", interactive=False, visible=False)

    # Store UI elements into the ui dict for external access if needed.
    ui.update(
        {
            # "gallery_accordion": gallery_accordion,
            "inference_accordion": inference_accordion,
            "start_frame": start_frame,
            "stop_frame": stop_frame,
            "model_dropdown": model_dropdown,
            "init_tracker_btn": init_tracker_btn,
            "tracking_btn": tracking_btn,
            "progress_text": progress_text,
            "display_mode_text": display_mode_text,
            "display_middle_result_btn": display_middle_result_btn,
            "cancel_btn": cancel_btn,
            "display": display,
        }
    )

    tracking_config = [start_frame, stop_frame, model_dropdown]
    # Set up the gallery from the label data.
    track_tab.select(
        fn=read_label_to_gallery,
        inputs=[storage_path, project_name, source_video],
        outputs=[label_list_state],
    )

    init_tracker_inputs = [storage_path, project_name, source_video] + tracking_config
    init_tracker_btn.click(
        fn=init_inference_tracker,
        inputs=init_tracker_inputs,
        outputs=tracker_state,
    )

    tracking_btn.click(
        fn=run_inference_tracker,
        inputs=tracker_state,
        outputs=progress_text,
    )

    display_middle_result_btn.click(
        fn=click_middle_result,
        inputs=tracker_state,
        outputs=[display, display_mode_text],
    )

    cancel_btn.click(fn=set_cancel, inputs=tracker_state)

    return ui



import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np

from ...utils.math import generate_permute_matrix
from ...utils.image import one_hot_mask

from ..layers.basic import seq_to_2d


class AOTEngine(nn.Module):
    def __init__(self,
                 aot_model,
                 device,
                 long_term_mem_gap=9999,
                 short_term_mem_skip=1,
                 max_len_long_term=9999):
        super().__init__()

        self.cfg = aot_model.cfg
        self.align_corners = aot_model.cfg.MODEL_ALIGN_CORNERS
        self.AOT = aot_model

        self.max_obj_num = aot_model.max_obj_num
        self.device = device
        self.long_term_mem_gap = long_term_mem_gap
        self.short_term_mem_skip = short_term_mem_skip
        self.max_len_long_term = max_len_long_term
        self.losses = None

        self.restart_engine()

    def forward(self,
                all_frames,
                all_masks,
                batch_size,
                obj_nums,
                step=0,
                tf_board=False,
                use_prev_pred=False,
                enable_prev_frame=False,
                use_prev_prob=False):  # only used for training
        if self.losses is None:
            self._init_losses()

        self.freeze_id = True if use_prev_pred else False
        aux_weight = self.aux_weight * max(self.aux_step - step, 0.) / self.aux_step

        self.offline_encoder(all_frames, all_masks)

        self.add_reference_frame(frame_step=0, obj_nums=obj_nums)

        grad_state = torch.no_grad if aux_weight == 0 else torch.enable_grad
        with grad_state():
            ref_aux_loss, ref_aux_mask = self.generate_loss_mask(
                self.offline_masks[self.frame_step], step)

        aux_losses = [ref_aux_loss]
        aux_masks = [ref_aux_mask]

        curr_losses, curr_masks = [], []
        if enable_prev_frame:
            self.set_prev_frame(frame_step=1)
            with grad_state():
                prev_aux_loss, prev_aux_mask = self.generate_loss_mask(
                    self.offline_masks[self.frame_step], step)
            aux_losses.append(prev_aux_loss)
            aux_masks.append(prev_aux_mask)
        else:
            self.match_propogate_one_frame()
            curr_loss, curr_mask, curr_prob = self.generate_loss_mask(
                self.offline_masks[self.frame_step], step, return_prob=True)
            self.update_short_term_memory(
                curr_mask if not use_prev_prob else curr_prob,
                None if use_prev_pred else self.assign_identity(
                    self.offline_one_hot_masks[self.frame_step]))
            curr_losses.append(curr_loss)
            curr_masks.append(curr_mask)

        self.match_propogate_one_frame()
        curr_loss, curr_mask, curr_prob = self.generate_loss_mask(
            self.offline_masks[self.frame_step], step, return_prob=True)
        curr_losses.append(curr_loss)
        curr_masks.append(curr_mask)
        for _ in range(self.total_offline_frame_num - 3):
            self.update_short_term_memory(
                curr_mask if not use_prev_prob else curr_prob,
                None if use_prev_pred else self.assign_identity(
                    self.offline_one_hot_masks[self.frame_step]))
            self.match_propogate_one_frame()
            curr_loss, curr_mask, curr_prob = self.generate_loss_mask(
                self.offline_masks[self.frame_step], step, return_prob=True)
            curr_losses.append(curr_loss)
            curr_masks.append(curr_mask)

        aux_loss = torch.cat(aux_losses, dim=0).mean(dim=0)
        pred_loss = torch.cat(curr_losses, dim=0).mean(dim=0)

        loss = aux_weight * aux_loss + pred_loss

        all_pred_mask = aux_masks + curr_masks

        all_frame_loss = aux_losses + curr_losses

        boards = {'image': {}, 'scalar': {}}

        return loss, all_pred_mask, all_frame_loss, boards

    def _init_losses(self):
        cfg = self.cfg

        from networks.layers.loss import CrossEntropyLoss, SoftJaccordLoss
        bce_loss = CrossEntropyLoss(
            cfg.TRAIN_TOP_K_PERCENT_PIXELS,
            cfg.TRAIN_HARD_MINING_RATIO * cfg.TRAIN_TOTAL_STEPS)
        iou_loss = SoftJaccordLoss()

        losses = [bce_loss, iou_loss]
        loss_weights = [0.5, 0.5]

        self.losses = nn.ModuleList(losses)
        self.loss_weights = loss_weights
        self.aux_weight = cfg.TRAIN_AUX_LOSS_WEIGHT
        self.aux_step = cfg.TRAIN_TOTAL_STEPS * cfg.TRAIN_AUX_LOSS_RATIO + 1e-5

    def encode_one_img_mask(self, img=None, mask=None, frame_step=-1):
        if frame_step == -1:
            frame_step = self.frame_step

        if self.enable_offline_enc:
            curr_enc_embs = self.offline_enc_embs[frame_step]
        elif img is None:
            curr_enc_embs = None
        else:
            curr_enc_embs = self.AOT.encode_image(img)

        if mask is not None:
            curr_one_hot_mask = one_hot_mask(mask, self.max_obj_num)
        elif self.enable_offline_enc:
            curr_one_hot_mask = self.offline_one_hot_masks[frame_step]
        else:
            curr_one_hot_mask = None

        return curr_enc_embs, curr_one_hot_mask

    def offline_encoder(self, all_frames, all_masks=None):
        self.enable_offline_enc = True
        self.offline_frames = all_frames.size(0) // self.batch_size

        # extract backbone features
        self.offline_enc_embs = self.split_frames(
            self.AOT.encode_image(all_frames), self.batch_size)
        self.total_offline_frame_num = len(self.offline_enc_embs)

        if all_masks is not None:
            # extract mask embeddings
            offline_one_hot_masks = one_hot_mask(all_masks, self.max_obj_num)
            self.offline_masks = list(
                torch.split(all_masks, self.batch_size, dim=0))
            self.offline_one_hot_masks = list(
                torch.split(offline_one_hot_masks, self.batch_size, dim=0))

        if self.input_size_2d is None:
            self.update_size(all_frames.size()[2:],
                             self.offline_enc_embs[0][-1].size()[2:])

    def assign_identity(self, one_hot_mask):
        if self.enable_id_shuffle:
            one_hot_mask = torch.einsum('bohw,bot->bthw', one_hot_mask,
                                        self.id_shuffle_matrix)

        id_emb = self.AOT.get_id_emb(one_hot_mask).view(
            self.batch_size, -1, self.enc_hw).permute(2, 0, 1)

        if self.training and self.freeze_id:
            id_emb = id_emb.detach()

        return id_emb

    def split_frames(self, xs, chunk_size):
        new_xs = []
        for x in xs:
            all_x = list(torch.split(x, chunk_size, dim=0))
            new_xs.append(all_x)
        return list(zip(*new_xs))

    def add_reference_frame(self,
                            img=None,
                            mask=None,
                            frame_step=-1,
                            obj_nums=None,
                            img_embs=None):
        # print("Run add_reference_frame at aot engine")
        if self.obj_nums is None and obj_nums is None:
            print('No objects for reference frame!')
            exit()
        elif obj_nums is not None:
            self.obj_nums = obj_nums

        if frame_step == -1:
            frame_step = self.frame_step

        if img_embs is None:
            curr_enc_embs, curr_one_hot_mask = self.encode_one_img_mask(
                img, mask, frame_step)
        else:
            _, curr_one_hot_mask = self.encode_one_img_mask(
                None, mask, frame_step)
            curr_enc_embs = img_embs

        if curr_enc_embs is None:
            print('No image for reference frame!')
            exit()

        if curr_one_hot_mask is None:
            print('No mask for reference frame!')
            exit()

        if self.input_size_2d is None:
            self.update_size(img.size()[2:], curr_enc_embs[-1].size()[2:])

        self.curr_enc_embs = curr_enc_embs
        self.curr_one_hot_mask = curr_one_hot_mask

        if self.pos_emb is None:
            self.pos_emb = self.AOT.get_pos_emb(curr_enc_embs[-1]).expand(
                self.batch_size, -1, -1,
                -1).view(self.batch_size, -1, self.enc_hw).permute(2, 0, 1)

        curr_id_emb = self.assign_identity(curr_one_hot_mask)
        self.curr_id_embs = curr_id_emb

        # self matching and propagation
        self.curr_lstt_output = self.AOT.LSTT_forward(curr_enc_embs,
                                                      None,
                                                      None,
                                                      curr_id_emb,
                                                      pos_emb=self.pos_emb,
                                                      size_2d=self.enc_size_2d)

        _, _, lstt_long_memories, lstt_short_memories = self.curr_lstt_output

        if self.long_term_memories is None:
            self.long_term_memories = lstt_long_memories
        else:
            self.update_long_term_memory(lstt_long_memories)

        self.last_mem_step = self.frame_step

        self.short_term_memories_list = [lstt_short_memories]
        self.short_term_memories = lstt_short_memories

    def set_prev_frame(self, img=None, mask=None, frame_step=1):
        self.frame_step = frame_step
        curr_enc_embs, curr_one_hot_mask = self.encode_one_img_mask(
            img, mask, frame_step)

        if curr_enc_embs is None:
            print('No image for previous frame!')
            exit()

        if curr_one_hot_mask is None:
            print('No mask for previous frame!')
            exit()

        self.curr_enc_embs = curr_enc_embs
        self.curr_one_hot_mask = curr_one_hot_mask

        curr_id_emb = self.assign_identity(curr_one_hot_mask)
        self.curr_id_embs = curr_id_emb

        # self matching and propagation
        self.curr_lstt_output = self.AOT.LSTT_forward(curr_enc_embs,
                                                      None,
                                                      None,
                                                      curr_id_emb,
                                                      pos_emb=self.pos_emb,
                                                      size_2d=self.enc_size_2d)

        _, _, lstt_long_memories, lstt_short_memories = self.curr_lstt_output

        if self.long_term_memories is None:
            self.long_term_memories = lstt_long_memories
        else:
            self.update_long_term_memory(lstt_long_memories)
        self.last_mem_step = frame_step

        self.short_term_memories_list = [lstt_short_memories]
        self.short_term_memories = lstt_short_memories

    def update_long_term_memory(self, new_long_term_memories):
        
        TOKEN_NUM = new_long_term_memories[0][0].shape[0]
        
        if self.long_term_memories is None:
            self.long_term_memories = new_long_term_memories
            return
        
        updated_long_term_memories = []
        for new_long_term_memory, last_long_term_memory in zip(
                new_long_term_memories, self.long_term_memories):
            updated_e = []
            for new_e, last_e in zip(new_long_term_memory, last_long_term_memory):
                if new_e is None or last_e is None:
                    updated_e.append(None)
                else:
                    if last_e.shape[0] >= self.max_len_long_term * TOKEN_NUM:
                        last_e = last_e[:(self.max_len_long_term - 1) * TOKEN_NUM]

                    updated_e.append(torch.cat([new_e, last_e], dim=0))
            updated_long_term_memories.append(updated_e)
        self.long_term_memories = updated_long_term_memories
        # print('llm status', self.long_term_memories[0][0][:3,0,:3])

    def update_short_term_memory(self, curr_mask, curr_id_emb=None, skip_long_term_update=False):
        if curr_id_emb is None:
            if len(curr_mask.size()) == 3 or curr_mask.size()[0] == 1:
                curr_one_hot_mask = one_hot_mask(curr_mask, self.max_obj_num)
            else:
                curr_one_hot_mask = curr_mask
            curr_id_emb = self.assign_identity(curr_one_hot_mask)

        lstt_curr_memories = self.curr_lstt_output[1]
        lstt_curr_memories_2d = []
        for layer_idx in range(len(lstt_curr_memories)):
            curr_k, curr_v = lstt_curr_memories[layer_idx][
                0], lstt_curr_memories[layer_idx][1]
            curr_k, curr_v = self.AOT.LSTT.layers[layer_idx].fuse_key_value_id(
                curr_k, curr_v, curr_id_emb)
            lstt_curr_memories[layer_idx][0], lstt_curr_memories[layer_idx][
                1] = curr_k, curr_v
            lstt_curr_memories_2d.append([
                seq_to_2d(lstt_curr_memories[layer_idx][0], self.enc_size_2d),
                seq_to_2d(lstt_curr_memories[layer_idx][1], self.enc_size_2d)
            ])

        self.short_term_memories_list.append(lstt_curr_memories_2d)
        self.short_term_memories_list = self.short_term_memories_list[
            -self.short_term_mem_skip:]
        self.short_term_memories = self.short_term_memories_list[0]

        if self.frame_step - self.last_mem_step >= self.long_term_mem_gap:
            # skip the update of long-term memory or not
            if not skip_long_term_update: 
                self.update_long_term_memory(lstt_curr_memories)
            self.last_mem_step = self.frame_step

    def match_propogate_one_frame(self, img=None, img_embs=None):
        self.frame_step += 1
        if img_embs is None:
            curr_enc_embs, _ = self.encode_one_img_mask(
                img, None, self.frame_step)
        else:
            curr_enc_embs = img_embs
        self.curr_enc_embs = curr_enc_embs

        self.curr_lstt_output = self.AOT.LSTT_forward(curr_enc_embs,
                                                      self.long_term_memories,
                                                      self.short_term_memories,
                                                      None,
                                                      pos_emb=self.pos_emb,
                                                      size_2d=self.enc_size_2d)

    def decode_current_logits(self, output_size=None):
        curr_enc_embs = self.curr_enc_embs
        curr_lstt_embs = self.curr_lstt_output[0]

        pred_id_logits = self.AOT.decode_id_logits(curr_lstt_embs,
                                                   curr_enc_embs)

        if self.enable_id_shuffle:  # reverse shuffle
            pred_id_logits = torch.einsum('bohw,bto->bthw', pred_id_logits,
                                          self.id_shuffle_matrix)

        # remove unused identities
        for batch_idx, obj_num in enumerate(self.obj_nums):
            pred_id_logits[batch_idx, (obj_num+1):] = - \
                1e+10 if pred_id_logits.dtype == torch.float32 else -1e+4

        self.pred_id_logits = pred_id_logits

        if output_size is not None:
            pred_id_logits = F.interpolate(pred_id_logits,
                                           size=output_size,
                                           mode="bilinear",
                                           align_corners=self.align_corners)

        return pred_id_logits

    def predict_current_mask(self, output_size=None, return_prob=False):
        if output_size is None:
            output_size = self.input_size_2d

        pred_id_logits = F.interpolate(self.pred_id_logits,
                                       size=output_size,
                                       mode="bilinear",
                                       align_corners=self.align_corners)
        pred_mask = torch.argmax(pred_id_logits, dim=1)

        if not return_prob:
            return pred_mask
        else:
            pred_prob = torch.softmax(pred_id_logits, dim=1)
            return pred_mask, pred_prob

    def calculate_current_loss(self, gt_mask, step):
        pred_id_logits = self.pred_id_logits

        pred_id_logits = F.interpolate(pred_id_logits,
                                       size=gt_mask.size()[-2:],
                                       mode="bilinear",
                                       align_corners=self.align_corners)

        label_list = []
        logit_list = []
        for batch_idx, obj_num in enumerate(self.obj_nums):
            now_label = gt_mask[batch_idx].long()
            now_logit = pred_id_logits[batch_idx, :(obj_num + 1)].unsqueeze(0)
            label_list.append(now_label.long())
            logit_list.append(now_logit)

        total_loss = 0
        for loss, loss_weight in zip(self.losses, self.loss_weights):
            total_loss = total_loss + loss_weight * \
                loss(logit_list, label_list, step)

        return total_loss

    def generate_loss_mask(self, gt_mask, step, return_prob=False):
        self.decode_current_logits()
        loss = self.calculate_current_loss(gt_mask, step)
        if return_prob:
            mask, prob = self.predict_current_mask(return_prob=True)
            return loss, mask, prob
        else:
            mask = self.predict_current_mask()
            return loss, mask

    def keep_gt_mask(self, pred_mask, keep_prob=0.2):
        pred_mask = pred_mask.float()
        gt_mask = self.offline_masks[self.frame_step].float().squeeze(1)

        shape = [1 for _ in range(pred_mask.ndim)]
        shape[0] = self.batch_size
        random_tensor = keep_prob + torch.rand(
            shape, dtype=pred_mask.dtype, device=pred_mask.device)
        random_tensor.floor_()  # binarize

        pred_mask = pred_mask * (1 - random_tensor) + gt_mask * random_tensor

        return pred_mask

    def restart_engine(self, batch_size=1, enable_id_shuffle=False):

        self.batch_size = batch_size
        self.frame_step = 0
        self.last_mem_step = -1
        self.enable_id_shuffle = enable_id_shuffle
        self.freeze_id = False

        self.obj_nums = None
        self.pos_emb = None
        self.enc_size_2d = None
        self.enc_hw = None
        self.input_size_2d = None

        self.long_term_memories = None
        self.short_term_memories_list = []
        self.short_term_memories = None

        self.enable_offline_enc = False
        self.offline_enc_embs = None
        self.offline_one_hot_masks = None
        self.offline_frames = -1
        self.total_offline_frame_num = 0

        self.curr_enc_embs = None
        self.curr_memories = None
        self.curr_id_embs = None

        if enable_id_shuffle:
            self.id_shuffle_matrix = generate_permute_matrix(
                self.max_obj_num + 1, batch_size, device=self.device)
        else:
            self.id_shuffle_matrix = None

    def update_size(self, input_size, enc_size):
        self.input_size_2d = input_size
        self.enc_size_2d = enc_size
        self.enc_hw = self.enc_size_2d[0] * self.enc_size_2d[1]


class AOTInferEngine(nn.Module):
    def __init__(self,
                 aot_model,
                 device,
                 long_term_mem_gap=9999,
                 short_term_mem_skip=1,
                 max_aot_obj_num=None,
                 max_len_long_term=9999,):
        super().__init__()

        self.cfg = aot_model.cfg
        self.AOT = aot_model

        if max_aot_obj_num is None or max_aot_obj_num > aot_model.max_obj_num:
            self.max_aot_obj_num = aot_model.max_obj_num
        else:
            self.max_aot_obj_num = max_aot_obj_num

        self.device = device
        self.long_term_mem_gap = long_term_mem_gap
        self.short_term_mem_skip = short_term_mem_skip
        self.max_len_long_term = max_len_long_term
        self.aot_engines = []

        self.restart_engine()
    def restart_engine(self):
        del (self.aot_engines)
        self.aot_engines = []
        self.obj_nums = None

    def separate_mask(self, mask, obj_nums):
        if mask is None:
            return [None] * len(self.aot_engines)
        if len(self.aot_engines) == 1:
            return [mask], [obj_nums]

        separated_obj_nums = [
            self.max_aot_obj_num for _ in range(len(self.aot_engines))
        ]
        if obj_nums % self.max_aot_obj_num > 0:
            separated_obj_nums[-1] = obj_nums % self.max_aot_obj_num

        if len(mask.size()) == 3 or mask.size()[0] == 1:
            separated_masks = []
            for idx in range(len(self.aot_engines)):
                start_id = idx * self.max_aot_obj_num + 1
                end_id = (idx + 1) * self.max_aot_obj_num
                fg_mask = ((mask >= start_id) & (mask <= end_id)).float()
                separated_mask = (fg_mask * mask - start_id + 1) * fg_mask
                separated_masks.append(separated_mask)
            return separated_masks, separated_obj_nums
        else:
            prob = mask
            separated_probs = []
            for idx in range(len(self.aot_engines)):
                start_id = idx * self.max_aot_obj_num + 1
                end_id = (idx + 1) * self.max_aot_obj_num
                fg_prob = prob[start_id:(end_id + 1)]
                bg_prob = 1. - torch.sum(fg_prob, dim=1, keepdim=True)
                separated_probs.append(torch.cat([bg_prob, fg_prob], dim=1))
            return separated_probs, separated_obj_nums

    def min_logit_aggregation(self, all_logits):
        if len(all_logits) == 1:
            return all_logits[0]

        fg_logits = []
        bg_logits = []

        for logit in all_logits:
            bg_logits.append(logit[:, 0:1])
            fg_logits.append(logit[:, 1:1 + self.max_aot_obj_num])

        bg_logit, _ = torch.min(torch.cat(bg_logits, dim=1),
                                dim=1,
                                keepdim=True)
        merged_logit = torch.cat([bg_logit] + fg_logits, dim=1)

        return merged_logit

    def soft_logit_aggregation(self, all_logits):
        if len(all_logits) == 1:
            return all_logits[0]

        fg_probs = []
        bg_probs = []

        for logit in all_logits:
            prob = torch.softmax(logit, dim=1)
            bg_probs.append(prob[:, 0:1])
            fg_probs.append(prob[:, 1:1 + self.max_aot_obj_num])

        bg_prob = torch.prod(torch.cat(bg_probs, dim=1), dim=1, keepdim=True)
        merged_prob = torch.cat([bg_prob] + fg_probs,
                                dim=1).clamp(1e-5, 1 - 1e-5)
        merged_logit = torch.logit(merged_prob)

        return merged_logit

    def add_reference_frame(self, img, mask, obj_nums, frame_step=-1):
        if isinstance(obj_nums, list):
            obj_nums = obj_nums[0]
        self.obj_nums = obj_nums
        aot_num = max(np.ceil(obj_nums / self.max_aot_obj_num), 1)
        while (aot_num > len(self.aot_engines)):
            new_engine = AOTEngine(self.AOT, self.device,
                                   self.long_term_mem_gap,
                                   self.short_term_mem_skip,
                                   self.max_len_long_term,)
            new_engine.eval()
            self.aot_engines.append(new_engine)

        separated_masks, separated_obj_nums = self.separate_mask(
            mask, obj_nums)
        img_embs = None
        for aot_engine, separated_mask, separated_obj_num in zip(
                self.aot_engines, separated_masks, separated_obj_nums):
            aot_engine.add_reference_frame(img,
                                        separated_mask,
                                        obj_nums=[separated_obj_num],
                                        frame_step=frame_step,
                                        img_embs=img_embs)
                
            if img_embs is None:  # reuse image embeddings
                img_embs = aot_engine.curr_enc_embs

        self.update_size()

    def match_propogate_one_frame(self, img=None):
        img_embs = None
        for aot_engine in self.aot_engines:
            aot_engine.match_propogate_one_frame(img, img_embs=img_embs)
            if img_embs is None:  # reuse image embeddings
                img_embs = aot_engine.curr_enc_embs

    def decode_current_logits(self, output_size=None):
        all_logits = []
        for aot_engine in self.aot_engines:
            all_logits.append(aot_engine.decode_current_logits(output_size))
        pred_id_logits = self.soft_logit_aggregation(all_logits)
        return pred_id_logits

    def update_memory(self, curr_mask, skip_long_term_update=False):
        _curr_mask = F.interpolate(curr_mask,self.input_size_2d)
        separated_masks, _ = self.separate_mask(_curr_mask, self.obj_nums)
        for aot_engine, separated_mask in zip(self.aot_engines,
                                              separated_masks):
            aot_engine.update_short_term_memory(separated_mask, 
                                                skip_long_term_update=skip_long_term_update)

    def update_size(self):
        self.input_size_2d = self.aot_engines[0].input_size_2d
        self.enc_size_2d = self.aot_engines[0].enc_size_2d
        self.enc_hw = self.aot_engines[0].enc_hw



import os
from .default import DefaultModelConfig

class ModelConfig(DefaultModelConfig):
    def __init__(self):
        super().__init__()
        self.MODEL_NAME = 'AOTL'

        self.MODEL_LSTT_NUM = 3

        self.TRAIN_LONG_TERM_MEM_GAP = 2

        self.TEST_LONG_TERM_MEM_GAP = 5


class DefaultModelConfig():
    def __init__(self):
        self.MODEL_NAME = 'AOTDefault'

        self.MODEL_VOS = 'aot'
        self.MODEL_ENGINE = 'aotengine'
        self.MODEL_ALIGN_CORNERS = True
        self.MODEL_ENCODER = 'mobilenetv2'
        self.MODEL_ENCODER_PRETRAIN = './pretrain_models/mobilenet_v2-b0353104.pth'
        self.MODEL_ENCODER_DIM = [24, 32, 96, 1280]  # 4x, 8x, 16x, 16x
        self.MODEL_ENCODER_EMBEDDING_DIM = 256
        self.MODEL_DECODER_INTERMEDIATE_LSTT = True
        self.MODEL_FREEZE_BN = True
        self.MODEL_FREEZE_BACKBONE = False
        self.MODEL_MAX_OBJ_NUM = 10
        self.MODEL_SELF_HEADS = 8
        self.MODEL_ATT_HEADS = 8
        self.MODEL_LSTT_NUM = 1
        self.MODEL_EPSILON = 1e-5
        self.MODEL_USE_PREV_PROB = False

        self.TRAIN_LONG_TERM_MEM_GAP = 9999
        self.TRAIN_AUG_TYPE = 'v1'

        self.TEST_LONG_TERM_MEM_GAP = 9999

        self.TEST_SHORT_TERM_MEM_SKIP = 1



import torch
import torch.nn as nn
import torch.nn.functional as F
from castle.aot.networks.layers.basic import ConvGN


class FPNSegmentationHead(nn.Module):
    def __init__(self,
                 in_dim,
                 out_dim,
                 decode_intermediate_input=True,
                 hidden_dim=256,
                 shortcut_dims=[24, 32, 96, 1280],
                 align_corners=True):
        super().__init__()
        self.align_corners = align_corners

        self.decode_intermediate_input = decode_intermediate_input

        self.conv_in = ConvGN(in_dim, hidden_dim, 1)

        self.conv_16x = ConvGN(hidden_dim, hidden_dim, 3)
        self.conv_8x = ConvGN(hidden_dim, hidden_dim // 2, 3)
        self.conv_4x = ConvGN(hidden_dim // 2, hidden_dim // 2, 3)

        self.adapter_16x = nn.Conv2d(shortcut_dims[-2], hidden_dim, 1)
        self.adapter_8x = nn.Conv2d(shortcut_dims[-3], hidden_dim, 1)
        self.adapter_4x = nn.Conv2d(shortcut_dims[-4], hidden_dim // 2, 1)

        self.conv_out = nn.Conv2d(hidden_dim // 2, out_dim, 1)

        self._init_weight()

    def forward(self, inputs, shortcuts):

        if self.decode_intermediate_input:
            x = torch.cat(inputs, dim=1)
        else:
            x = inputs[-1]

        x = F.relu_(self.conv_in(x))
        x = F.relu_(self.conv_16x(self.adapter_16x(shortcuts[-2]) + x))

        x = F.interpolate(x,
                          size=shortcuts[-3].size()[-2:],
                          mode="bilinear",
                          align_corners=self.align_corners)
        x = F.relu_(self.conv_8x(self.adapter_8x(shortcuts[-3]) + x))

        x = F.interpolate(x,
                          size=shortcuts[-4].size()[-2:],
                          mode="bilinear",
                          align_corners=self.align_corners)
        x = F.relu_(self.conv_4x(self.adapter_4x(shortcuts[-4]) + x))

        x = self.conv_out(x)

        return x

    def _init_weight(self):
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)



from .aot_engine import AOTEngine, AOTInferEngine
from .deaot_engine import DeAOTEngine, DeAOTInferEngine


def build_engine(name, phase='train', **kwargs):
    if name == 'aotengine':
        if phase == 'train':
            return AOTEngine(**kwargs)
        elif phase == 'eval':
            return AOTInferEngine(**kwargs)
        else:
            raise NotImplementedError
    elif name == 'deaotengine':
        if phase == 'train':
            return DeAOTEngine(**kwargs)
        elif phase == 'eval':
            return DeAOTInferEngine(**kwargs)
        else:
            raise NotImplementedError
    else:
        raise NotImplementedError




from .download import download_file
from .video_io import ReadArray
from .video_align import get_mask
from tqdm import tqdm
import numpy as np
import gc

import torch
import torchvision.transforms as tt
from torch.utils.data import Dataset, DataLoader

import platform
OS_SYS = platform.uname().system
if OS_SYS == 'Darwin':
    DEFAULT_DEVICE = 'mps'
elif torch.cuda.is_available():
    DEFAULT_DEVICE = 'cuda'
else:
    DEFAULT_DEVICE = 'cpu'




resolution = 518
patch_len = resolution // 14
img2tensor = tt.Compose([
    tt.ToTensor(), # range [0, 255] -> [0.0,1.0]
    tt.Resize((resolution, resolution), antialias=True),
    tt.Normalize(mean=0.5, std=0.2), # range [0.0,1.0] -> [-2.5, 2.5]

])

img2mask = tt.Compose([
    tt.ToTensor(), # range [0, 255] -> [0.0,1.0]
    tt.Resize((resolution, resolution), antialias=True),
])

class DinoV2latentGen:
    def __init__(self, model_cfg):
        self.model = torch.hub.load('facebookresearch/dinov2', 'dinov2_vitb14_reg')
        self.device = model_cfg['device']
        self.n_feature = self.model.embed_dim
        print("Device: ", self.device)
        
    def batch_run(self, X):
        if type(X) == list:
            X = torch.stack(X)
        return self.run(X)
    
    def single_run(self, x):
        X = torch.unsqueeze(x, 0)
        return self.run(X)
        
    def run(self, X):
        self.model.eval()
        self.model.to(self.device)
        with torch.no_grad():
            X = X.to(self.device)
            result = self.model.forward_features(X)
        
        return result['x_norm_patchtokens'].detach().cpu().numpy()
    
    def __del__(self):
        del self.model
        
   




class ObserverDINOv2:
    def __init__(self, dinov2_args):

        self.model = DinoV2latentGen(dinov2_args)
        self.n_feature = self.model.n_feature
        self.batch_size = dinov2_args['batch_size']

    def nan_latent(self):
        return np.full((self.n_feature), np.nan, dtype=np.float16)

    def extract_image_latent(self, frame, mask, select_roi):
        return self.extract_batch_latent([frame], [mask], select_roi)[0]

    def extract_batch_latent(self, frame_list, mask_list, select_roi):
        batch_latent = []
        mask_roi_list = [img2mask(get_mask(it, select_roi)) for it in mask_list]
        frame_list = [img2tensor(it) for it in frame_list]

        patch_feature = self.model.batch_run(frame_list)

        for img, mask in zip(patch_feature, mask_roi_list):
            latent = img.reshape((patch_len, patch_len, self.n_feature))
            small_mask = mask.reshape(resolution//14, 14, resolution//14, 14).sum(axis=(1, 3))
            sum_mask = small_mask.sum()
            result = small_mask[:, :, np.newaxis] * latent
            latent_mask_ave = result.sum(axis=0).sum(axis=0) / sum_mask
            batch_latent.append(latent_mask_ave)

        return batch_latent

    def extract_video_latent(self, video_path, mask_video_path, roi_rgb, batch_size=16):
        # TODO
        pass

    def __del__(self):
        del self.model
        torch.cuda.empty_cache()


def download_dinov2_ckpt(model_type):
    if model_type == 'dinov2_vitb14_reg':
        ckpt_path = 'ckpt/dinov2_vitb14_reg4_pretrain.pth'
        download_file('https://dl.fbaipublicfiles.com/dinov2/dinov2_vitb14/dinov2_vitb14_reg4_pretrain.pth', ckpt_path)

        return ckpt_path
    else:
        assert False, f"model_type mismatch {model_type}, expect dinov2_vitb14_reg4_pretrain."



def generate_dinov2(model_type='dinov2_vitb14_reg', device='', batch_size=16):
    if len(device) == 0:
        device = DEFAULT_DEVICE
    dinov2_args = {
        "model_type": model_type,
        "device": device,
        "batch_size": batch_size,
    }
    return ObserverDINOv2(dinov2_args)


# import gradio as gr
# import os
# import json
# import glob
# import numpy as np

# def init_drop(storage_path, project_name):
#     if project_name == None:
#         return gr.update(choices=[])

#     project_path = os.path.join(storage_path, project_name)
#     project_config_path = os.path.join(project_path, 'config.json')
#     project_config = json.load(open(project_config_path, 'r'))
#     latent_list = [k for k, v in project_config['latent'].items()]
#     latent_list.append("All")


#     return gr.update(choices=latent_list), gr.update(choices=[]), latent_list, []



# def select_latent(target_latent,  unselect_list, select_list):
#     if target_latent == 'All':
#         unselect_list.remove('All')
#         select_list.extend(unselect_list)
#         unselect_list = ['All']
#     else:
#         unselect_list.remove(target_latent)
#         select_list.append(target_latent)
#     return gr.update(choices=unselect_list), gr.update(choices=select_list), unselect_list, select_list

# def unselect_latent(target_latent,  unselect_list, select_list):
#     select_list.remove(target_latent)
#     unselect_list.append(target_latent)
#     return gr.update(choices=unselect_list), gr.update(choices=select_list, value=""), unselect_list, select_list


# def build_umap_intput(storage_path, project_name, umap_input_name, select_list):
#     if len(umap_input_name) == 0:
#         gr.Info("The name is not allow empty.")
#         return
#     project_path = os.path.join(storage_path, project_name)
#     project_config_path = os.path.join(project_path, 'config.json')
#     project_config = json.load(open(project_config_path, 'r'))
#     latent_dir_path = os.path.join(storage_path, project_name, 'latent')
#     if not 'umap_input' in project_config:
#         project_config['umap_input'] = dict()

#     umap_input_dir_path = os.path.join(project_path, 'umap_input')
#     os.makedirs(umap_input_dir_path, exist_ok=True)
#     umap_input_path = os.path.join(umap_input_dir_path, f'{umap_input_name}_umap_input.npz')

#     umap_input_mask = []
#     for it in select_list:
#         latent_path = os.path.join(latent_dir_path, it)
#         latent = np.load(latent_path)['latent']
#         mask = (~np.isnan(latent[:, 0]))
#         umap_input_mask.extend(mask)

#     umap_input_mask = np.array(umap_input_mask)
#     # print('umap_input_mask', umap_input_mask.shape)

#     np.savez_compressed(umap_input_path, umap_mask=umap_input_mask)
#     project_config['umap_input'][f'{umap_input_name}_umap_input.npz'] = {
#         'latent': select_list,
#         'source': [project_config['latent'][it] for it in select_list],

#     }
#     json.dump(project_config, open(project_config_path,'w'))
#     gr.Info("Successful build the list as a UMAP Input")
#     return



# def create_cluster_input_ui(storage_path, project_name, cluster_page_tab):
#     ui = dict()

#     select_list = gr.State([])
#     unselect_list = gr.State([])
#     with gr.Row(visible=True):
#         with gr.Column(scale=5):
#             ui['unselect_drop'] = gr.Dropdown(label='Unselect List', interactive=True)
#             ui['select_latent'] = gr.Button('Select')
#         with gr.Column(scale=5):
#             ui['select_drop'] = gr.Dropdown(label='Select List', interactive=True)
#             ui['unselect_latent'] = gr.Button('Unselect')
#     ui['name'] = gr.Textbox(label='Name')
#     ui['build'] = gr.Button('Build List as UMAP Input')



#     cluster_page_tab.select(
#         fn=init_drop,
#         inputs=[storage_path, project_name],
#         outputs=[ui['unselect_drop'], ui['select_drop'], unselect_list, select_list]
#     )


#     ui['select_latent'].click(
#         fn=select_latent,
#         inputs=[ui['unselect_drop'], unselect_list, select_list],
#         outputs=[ui['unselect_drop'], ui['select_drop'], unselect_list, select_list],
#     )

#     ui['unselect_latent'].click(
#         fn=unselect_latent,
#         inputs=[ui['select_drop'], unselect_list, select_list],
#         outputs=[ui['unselect_drop'], ui['select_drop'], unselect_list, select_list],
#     )

#     ui['build'].click(
#         fn=build_umap_intput,
#         inputs=[storage_path, project_name, ui['name'], select_list],
#     )


#     return ui


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Patch

_palette = [
    '#7AE4F0', '#FFD0EC', '#FBC471', '#6EE368', '#C1B5EA', '#A7CCED',
    '#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A', '#19D3F3', 
    '#FF6692', '#B6E880', '#FF97FF', '#FECB52', '#FD3216', '#00FE35', 
    '#6A76FC', '#FED4C4', '#FE00CE', '#0DF9FF', '#F6F926', '#FF9616', 
    '#479B55', '#EEA6FB', '#DC587D', '#D626FF', '#6E899C', '#00B5F7', 
    '#B68E00', '#C9FBE5', '#FF0092', '#22FFA7', '#E3EE9E', '#86CE00', 
    '#BC7196', '#7E7DCD', '#FC6955', '#E48F72', '#66c5cc', '#f6cf71', 
    '#f89c74', '#dcb0f2', '#87c55f', '#9eb9f3', '#fe88b1', '#c9db74', 
    '#8be0a4', '#b497e7', '#b3b3b3', '#e58606', '#5d69b1', '#52bca3', 
    '#99c945', '#cc61b0', '#24796c', '#daa51b', '#2f8ac4', '#764e9f', 
    '#ed645a', '#a5aa99'
]


    
class Latent:
    def __init__(self, data, window=1):
        assert data.ndim == 2, 'Only design for (T x num_feats)'
        
        # Temporal concatenation
        n = (len(data) // window) * window
        num_feats = data.shape[-1]
        self.window = window
        self.data = data[:n].reshape((-1,  num_feats * window))
        

        self.syllables = np.zeros(len(self.data)).astype(int) - 1
        self.meta = []
        self.lookformeta = dict()
        
        # Init first cluster
        self.meta.append({
            'name': 'C0',
            'color': 'grey'
        })
        self.lookformeta['C0'] = 0
        self.syllables[~np.isnan(self.data.sum(axis=1))] = 0
        self.need_maintain_key_frames = True
        self.used_palette = set()
        
    
    def select(self, cid):
        # Generate FocusLatent
        export = np.copy(self.data)
        export[self.syllables != cid] = np.nan
        return FocusLatent(export, self.used_palette)
    
    def split(self, foucs_latent):
        self.need_maintain_key_frames = True
        focus = foucs_latent.focus
        focus_cluster = foucs_latent.cluster
        focus_cluster_set = set(focus_cluster[focus])
        if -1 in focus_cluster_set:
            focus_cluster_set.remove(-1)
        
        for it in focus_cluster_set:
            cid = len(self.meta)
            self.meta.append({
                'name': f'C{cid}',
                'color': foucs_latent.palette(it)
            })
            self.lookformeta[f'C{cid}'] = cid
            self.used_palette.add(foucs_latent.palette(it))
            mask = focus * (focus_cluster == it)
            self.syllables[mask] = cid
        
        
    def merge(self, cids):
        self.syllables[self.syllables == cid2] = cid1
        self.need_maintain_key_frames = True
        assert hasattr(self, 'syllables'), 'Do split first'
        assert len(cids) >= 2
        
        for i in range(1, len(cids)):
            self.syllables[self.syllables == cids[i]] = cids[0]
            
            
            
        
    def change_name(self, cid, name):
        assert not name in self.lookformeta, 'This name already be used.'
        self.meta[cid]['name'] = name
        return True
        
    def maintain_key_frames(self):
        self.need_maintain_key_frames = False
        n = len(self.data)
        self.key_frames = [0] + [i + 1 for i in range(n - 1) if self.syllables[i] != self.syllables[i + 1]] + [n - 1]
        
    def palette(self, c):
        if c >= 0 and c < len(self.meta):
            return self.meta[c]['color']
        else:
            return 'grey'
        
    def plot(self, legend=True):
        if self.need_maintain_key_frames:
            self.maintain_key_frames()

        widths, colors, lefts = [], [], []
        for j in range(len(self.key_frames) - 1):
            widths.append(self.key_frames[j + 1] - self.key_frames[j])
            colors.append(self.palette(self.syllables[self.key_frames[j]]))
            lefts.append(self.key_frames[j])

        # Plot the bar chart
        plt.bar(lefts, height=[1] * len(widths), width=widths, color=colors, align='edge', edgecolor='none')
        plt.xlim(0, self.key_frames[-1])
        plt.ylim(0, 1)
        plt.yticks([])

        # Prepare legend if necessary
        if legend:
            unique_categories = sorted(set(self.syllables[self.key_frames[j]] for j in range(len(self.key_frames) - 1)))
            if -1 in unique_categories:
                unique_categories.remove(-1)

            legend_handles = [
                Patch(color=self.palette(cat), label=self.meta[cat]['name'])
                for cat in unique_categories
            ]
            plt.legend(handles=legend_handles, title="Categories")


def gen_palette(avoid):
    res = [it for it in _palette if not it in avoid]
    return res

        
class FocusLatent:
    def __init__(self, data, color_avoid):
        self.data = data
        self.focus = (~np.isnan(self.data.sum(axis=1)))
        self.color_avoid = color_avoid
        self._palette = gen_palette(color_avoid)
        
        
    def palette(self, x):
        if x == -1:
            return 'grey'
        return self._palette[x % len(self._palette)]

        
    def gen_embedding(self, configs, device='cpu'):
        assert device == 'cpu' or device == 'gpu'
        if device == 'cpu':
            from umap import UMAP
        else:
            # from cuml.manifold import UMAP
            from myumap import UMAP
        
        if not type(configs) == list:
            configs = [configs]

        self.models = []
        Z = self.data[self.focus]
        for it in configs:
            model = UMAP(**it)
            Z = model.fit_transform(Z)
            self.models.append(model)

        self.embedding = np.zeros((len(self.data), Z.shape[-1])) + np.nan
        self.embedding[self.focus] = Z
        del UMAP
        
    def inference_embedding(self, data, device='cpu'):
        assert device == 'cpu' or device == 'gpu'
        if device == 'cpu':
            from umap import UMAP
        else:
            # from cuml.manifold import UMAP
            from myumap import UMAP

        Z = data
        for model in self.models:
            Z = model.transform(Z)
            
        del UMAP
        return Z
        
        
    def gen_cluster(self, config, device='cpu'):
        assert hasattr(self, 'embedding'), 'Do gen_embedding first'
        if device == 'cpu':
            from sklearn.cluster import DBSCAN
        else:
            from cuml.cluster import DBSCAN
            
        self.cluster = np.zeros(len(self.data)).astype(int) - 1
        C = DBSCAN(**config).fit_predict(self.embedding[self.focus])
        self.cluster[self.focus] = C
            
        
    def merge(self, cids):
        assert hasattr(self, 'cluster'), 'Do gen_cluster first'
        assert len(cids) >= 2
        
        for i in range(1, len(cids)):
            self.cluster[self.cluster == cids[i]] = cids[0]
        
        
    def plot(self, dimensions=[0, 1], legend=True):
        assert hasattr(self, 'embedding')
        assert len(dimensions) == 2
        if hasattr(self, 'cluster'):
            for it in range(0, self.cluster.max()+1):
                plt.scatter(x=self.embedding[self.cluster == it, dimensions[0]], 
                            y=self.embedding[self.cluster == it, dimensions[1]], 
                            c=self.palette(it), 
                            label=f'{it}')
            if -1 in self.cluster:
                
                plt.scatter(x=self.embedding[self.cluster == -1, dimensions[0]], 
                            y=self.embedding[self.cluster == -1, dimensions[1]], 
                            c='grey',
                            label=f'-1')
            if legend:
                plt.legend()
        else:
            plt.scatter(x=self.embedding[:, dimensions[0]], 
                        y=self.embedding[:, dimensions[1]], 
                        c='grey')


# Should be replace by explorer.py

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Patch
import platform
OS_SYS = platform.uname().system
import torch

if OS_SYS == 'Darwin':
    DEFAULT_DEVICE = 'mps'
elif torch.cuda.is_available():
    DEFAULT_DEVICE = 'cuda'
else:
    DEFAULT_DEVICE = 'cpu'


_palette = ['#636EFA', '#EF553B', '#00CC96', '#AB63FA', '#FFA15A', '#19D3F3', '#FF6692', '#B6E880', '#FF97FF', '#FECB52']
_palette += ['#FD3216', '#00FE35', '#6A76FC', '#FED4C4', '#FE00CE', '#0DF9FF', '#F6F926', '#FF9616', '#479B55', '#EEA6FB', '#DC587D', '#D626FF', '#6E899C', '#00B5F7', '#B68E00', '#C9FBE5', '#FF0092', '#22FFA7', '#E3EE9E', '#86CE00', '#BC7196', '#7E7DCD', '#FC6955', '#E48F72']
_palette += ['#66c5cc', '#f6cf71', '#f89c74', '#dcb0f2', '#87c55f', '#9eb9f3', '#fe88b1', '#c9db74', '#8be0a4', '#b497e7', '#b3b3b3']
_palette += ['#e58606', '#5d69b1', '#52bca3', '#99c945', '#cc61b0', '#24796c', '#daa51b', '#2f8ac4', '#764e9f', '#ed645a', '#a5aa99']





def generate_palette(avoid):
    res = [it for it in _palette if not it in avoid]
    return res
    



class Latent:
    def __init__(self, raw, time_window=1, device=''):
        if len(device) == 0:
            device = DEFAULT_DEVICE
        n = (len(raw) // time_window) * time_window
        num_feature = raw.shape[-1]
        self.time_window = time_window
        self.data = raw[:n].reshape((-1,  num_feature * time_window))
        print('self.data:', self.data.shape)
        self.cluster = np.zeros(len(self.data)).astype(int)
        self.cluster[np.isnan(self.data.sum(axis=1))] = -1
        self.cluster_meta = dict()
        self.behavior_name2cluster_id = dict()
        
        self.cluster_meta[0] = {
            'name': 'init',
            'color': 'grey'
        }
        self.behavior_name2cluster_id['init'] = 0
        self.num_cluster = 1
        self.need_maintain_key_frames = True
        self.device=device
        
        self.used_palette = set()
        
    def get_time_window(self):
        return self.time_window

    def select(self, selected_cluster):
        if type(selected_cluster) == str:
            selected_cluster = self.behavior_name2cluster_id[selected_cluster]
        return LocalLatent(self.data[self.cluster == selected_cluster], self.cluster == selected_cluster, color_avoid=self.used_palette, device=self.device)
    
    def merge(self, cluster_ids):
        cluster_ids = np.array(cluster_ids)
        mi = cluster_ids.min()

        for it in cluster_ids:
            self.cluster[self.cluster == it] = mi

        self.need_maintain_key_frames = True

    def maintain_key_frames(self):
        if hasattr(self, 'key_frames'):
            delattr(self, 'key_frames')
        n = len(self.data)
        self.key_frames = [0] + [i + 1 for i in range(n - 1) if self.cluster[i] != self.cluster[i + 1]] + [n - 1]
        self.need_maintain_key_frames = False

    def palette(self, c):
        if c in self.cluster_meta:
            return self.cluster_meta[c]['color']
        else:
            return 'grey'

    def plot_syllables(self):
        if self.need_maintain_key_frames:
            self.maintain_key_frames()
            

        widths = [self.key_frames[j+1] - self.key_frames[j] for j in range(len(self.key_frames)-1)]
        colors = [self.palette(self.cluster[self.key_frames[j]]) for j in range(len(self.key_frames)-1)]
        lefts = self.key_frames[:-1]


        
        plt.bar(lefts, height=[1]*len(widths), width=widths, color=colors, align='edge', edgecolor='none')
        plt.xlim(0, self.key_frames[-1])
        plt.ylim(0, 1)
        plt.yticks([])
        unique_categories = sorted(set(self.cluster[self.key_frames[j]] for j in range(len(self.key_frames)-1)))
        if -1 in unique_categories:
            unique_categories.remove(-1)

        legend_handles = [Patch(color=self.palette(cat), label=self.cluster_meta[cat]['name']) for cat in unique_categories]

        plt.legend(handles=legend_handles, title="Categories")



    def import_local_latent(self, local_latent):
        assert hasattr(local_latent, 'cluster')
        cluster = local_latent.cluster
        index_mask = local_latent.index_mask
        old_cluster = self.cluster[index_mask]

        # Check Name used?
        for _, it in local_latent.export.items():
            assert not it['name'] in self.behavior_name2cluster_id, 'new name be used'

        for cluster_local_id, it in local_latent.export.items():
            cluster_id = self.num_cluster
            self.num_cluster += 1

            old_cluster[cluster == cluster_local_id] = cluster_id
            self.cluster_meta[cluster_id] = {
                'name': it['name'],
                'color': it['color']
            }
            self.behavior_name2cluster_id[it['name']] = cluster_id
            self.used_palette.add(it['color'])

        self.cluster[index_mask] = old_cluster

        self.need_maintain_key_frames = True


class LocalLatent:
    def __init__(self, data, index_mask, color_avoid, device):
        self.data = data
        self.index_mask = index_mask
        self.device = device
        self.color_avoid = color_avoid
        self._palette = generate_palette(color_avoid)

        self.export = dict()
        

    def build_embedding(self, configs):
        if self.device == 'cpu' or self.device == 'mps':
            from umap import UMAP
        elif 'cuda' in self.device:
            try:
                # from cuml.manifold import UMAP
                from castle.utils.myumap import UMAP
            except:
                from umap import UMAP
        else:
            assert False, f'device error, expect cpu, mps, or cuda, got {self.device}'
        Z = self.data
        if hasattr(self, 'embedding'):
            delattr(self, 'embedding')

        if not type(configs) == list:
            configs = [configs]

        for it in configs:
            Z = UMAP(**it).fit_transform(Z)

        self.embedding = np.array(Z)



    def build_cluster(self, method, configs):
        if self.device == 'cpu':
            from sklearn.cluster import DBSCAN

        elif 'cuda' in self.device:
            try:
                from cuml.cluster import DBSCAN
            except:
                from sklearn.cluster import DBSCAN

            

        assert hasattr(self, 'embedding')
        if hasattr(self, 'cluster'):
            delattr(self, 'cluster')


        if method == 'dbscan':
            self.cluster = DBSCAN(**configs).fit_predict(self.embedding)
        else:
            assert False, f"method name should be dbscan, but got {method}."

    def palette(self, x):
        if x == -1:
            return '#DDDDDD'
        return self._palette[x % len(self._palette)]

    
    def plot_embedding(self, dims=[0, 1]):
        assert hasattr(self, 'embedding')
        assert len(dims) == 2, 'dims should'
        if hasattr(self, 'cluster'):
            for it in range(0, self.cluster.max()+1):
                plt.scatter(x=self.embedding[self.cluster == it, dims[0]], 
                            y=self.embedding[self.cluster == it, dims[1]], 
                            c=self.palette(it), 
                            label=f'{it}')
            if -1 in self.cluster:
                plt.scatter(x=self.embedding[self.cluster == -1, dims[0]], 
                            y=self.embedding[self.cluster == -1, dims[1]], 
                            c='grey',
                            label=f'-1')
            plt.legend()
        else:
            plt.scatter(x=self.embedding[:, dims[0]], 
                        y=self.embedding[:, dims[1]], 
                        c='grey')
            
    def merge(self, cluster_ids):
        cluster_ids = np.array(cluster_ids)
        mi = cluster_ids.min()

        for it in cluster_ids:
            self.cluster[self.cluster == it] = mi


    def label_cluster(self, cluster_id, cluster_name, cluster_color=''):
        tmp = dict()
        tmp['name'] = cluster_name
        tmp['color'] = cluster_color if len(cluster_color) > 0 else self._palette[cluster_id]
        # tmp['data'] = self.cluster == cluster_id

        self.export[cluster_id] = tmp
    
    def clean_label(self):
        self.export = dict()





import os
import h5py


class H5IO:
    def __init__(self, file_path):
        self.file_path = file_path
        self.config = dict()
        self.reset()

    def __setitem__(self, index, mask):
        self.write_mask(index, mask)

    def write_mask(self, index, mask):
        self.check()
        if str(index) in self.f:
            dset = self.f[str(index)] # for Overwrite previous results
            dset[:] = mask
        else:
            dset = self.f.create_dataset(str(index), mask.shape, dtype='uint8', compression="gzip", compression_opts=3)
            dset[:] = mask

    def __getitem__(self, index):
        return self.read_mask(index)
    
    def has_mask(self, index):
        return str(index) in self.f

    def read_mask(self, index):
        assert str(index) in self.f, f"Without mask at frame {index}"
        return self.f[str(index)][:]


    def read_config(self, key):
        value = self.f[key][()]
        print('read_config', key, value)
        return value


    def write_config(self, key, value):
        print('write_config', key, value)
        if key in self.f:
            del self.f[key]

        self.f.create_dataset(key, data=value)


    def check(self):
        self.reset_count += 1
        if self.reset_count > 5000:
            self.reset()

    def reset(self):
        self.reset_count = 0
        if hasattr(self, 'f'):
            self.f.close()

        mode = 'a' if os.path.isfile(self.file_path) else 'w'
        self.f = h5py.File(self.file_path, mode)

    def get_n_rois(self):
        return int(self.f['n_rois'][()])
    
    def __len__(self):
        return int(self.f['total_frames'][()])
    


    def __del__(self):
        if hasattr(self, 'f') and self.f.id.valid:
            self.f.close()
            del self.f





from __future__ import division
import os
import shutil
import json
import cv2
from PIL import Image

import numpy as np
from torch.utils.data import Dataset

from utils.image import _palette


class VOSTest(Dataset):
    def __init__(self,
                 image_root,
                 label_root,
                 seq_name,
                 images,
                 labels,
                 rgb=True,
                 transform=None,
                 single_obj=False,
                 resolution=None):
        self.image_root = image_root
        self.label_root = label_root
        self.seq_name = seq_name
        self.images = images
        self.labels = labels
        self.obj_num = 1
        self.num_frame = len(self.images)
        self.transform = transform
        self.rgb = rgb
        self.single_obj = single_obj
        self.resolution = resolution

        self.obj_nums = []
        self.obj_indices = []

        curr_objs = [0]
        for img_name in self.images:
            self.obj_nums.append(len(curr_objs) - 1)
            current_label_name = img_name.split('.')[0] + '.png'
            if current_label_name in self.labels:
                current_label = self.read_label(current_label_name)
                curr_obj = list(np.unique(current_label))
                for obj_idx in curr_obj:
                    if obj_idx not in curr_objs:
                        curr_objs.append(obj_idx)
            self.obj_indices.append(curr_objs.copy())

        self.obj_nums[0] = self.obj_nums[1]

    def __len__(self):
        return len(self.images)

    def read_image(self, idx):
        img_name = self.images[idx]
        img_path = os.path.join(self.image_root, self.seq_name, img_name)
        img = cv2.imread(img_path)
        img = np.array(img, dtype=np.float32)
        if self.rgb:
            img = img[:, :, [2, 1, 0]]
        return img

    def read_label(self, label_name, squeeze_idx=None):
        label_path = os.path.join(self.label_root, self.seq_name, label_name)
        label = Image.open(label_path)
        label = np.array(label, dtype=np.uint8)
        if self.single_obj:
            label = (label > 0).astype(np.uint8)
        elif squeeze_idx is not None:
            squeezed_label = label * 0
            for idx in range(len(squeeze_idx)):
                obj_id = squeeze_idx[idx]
                if obj_id == 0:
                    continue
                mask = label == obj_id
                squeezed_label += (mask * idx).astype(np.uint8)
            label = squeezed_label
        return label

    def __getitem__(self, idx):
        img_name = self.images[idx]
        current_img = self.read_image(idx)
        height, width, channels = current_img.shape
        if self.resolution is not None:
            width = int(np.ceil(
                float(width) * self.resolution / float(height)))
            height = int(self.resolution)

        current_label_name = img_name.split('.')[0] + '.png'
        obj_num = self.obj_nums[idx]
        obj_idx = self.obj_indices[idx]

        if current_label_name in self.labels:
            current_label = self.read_label(current_label_name, obj_idx)
            sample = {
                'current_img': current_img,
                'current_label': current_label
            }
        else:
            sample = {'current_img': current_img}

        sample['meta'] = {
            'seq_name': self.seq_name,
            'frame_num': self.num_frame,
            'obj_num': obj_num,
            'current_name': img_name,
            'height': height,
            'width': width,
            'flip': False,
            'obj_idx': obj_idx
        }

        if self.transform is not None:
            sample = self.transform(sample)
        return sample


class YOUTUBEVOS_Test(object):
    def __init__(self,
                 root='./datasets/YTB',
                 year=2018,
                 split='val',
                 transform=None,
                 rgb=True,
                 result_root=None):
        if split == 'val':
            split = 'valid'
        root = os.path.join(root, str(year), split)
        self.db_root_dir = root
        self.result_root = result_root
        self.rgb = rgb
        self.transform = transform
        self.seq_list_file = os.path.join(self.db_root_dir, 'meta.json')
        self._check_preprocess()
        self.seqs = list(self.ann_f.keys())
        self.image_root = os.path.join(root, 'JPEGImages')
        self.label_root = os.path.join(root, 'Annotations')

    def __len__(self):
        return len(self.seqs)

    def __getitem__(self, idx):
        seq_name = self.seqs[idx]
        data = self.ann_f[seq_name]['objects']
        obj_names = list(data.keys())
        images = []
        labels = []
        for obj_n in obj_names:
            images += map(lambda x: x + '.jpg', list(data[obj_n]["frames"]))
            labels.append(data[obj_n]["frames"][0] + '.png')
        images = np.sort(np.unique(images))
        labels = np.sort(np.unique(labels))

        try:
            if not os.path.isfile(
                    os.path.join(self.result_root, seq_name, labels[0])):
                if not os.path.exists(os.path.join(self.result_root,
                                                   seq_name)):
                    os.makedirs(os.path.join(self.result_root, seq_name))
                shutil.copy(
                    os.path.join(self.label_root, seq_name, labels[0]),
                    os.path.join(self.result_root, seq_name, labels[0]))
        except Exception as inst:
            print(inst)
            print('Failed to create a result folder for sequence {}.'.format(
                seq_name))

        seq_dataset = VOSTest(self.image_root,
                              self.label_root,
                              seq_name,
                              images,
                              labels,
                              transform=self.transform,
                              rgb=self.rgb)
        return seq_dataset

    def _check_preprocess(self):
        _seq_list_file = self.seq_list_file
        if not os.path.isfile(_seq_list_file):
            print(_seq_list_file)
            return False
        else:
            self.ann_f = json.load(open(self.seq_list_file, 'r'))['videos']
            return True


class YOUTUBEVOS_DenseTest(object):
    def __init__(self,
                 root='./datasets/YTB',
                 year=2018,
                 split='val',
                 transform=None,
                 rgb=True,
                 result_root=None):
        if split == 'val':
            split = 'valid'
        root_sparse = os.path.join(root, str(year), split)
        root_dense = root_sparse + '_all_frames'
        self.db_root_dir = root_dense
        self.result_root = result_root
        self.rgb = rgb
        self.transform = transform
        self.seq_list_file = os.path.join(root_sparse, 'meta.json')
        self._check_preprocess()
        self.seqs = list(self.ann_f.keys())
        self.image_root = os.path.join(root_dense, 'JPEGImages')
        self.label_root = os.path.join(root_sparse, 'Annotations')

    def __len__(self):
        return len(self.seqs)

    def __getitem__(self, idx):
        seq_name = self.seqs[idx]

        data = self.ann_f[seq_name]['objects']
        obj_names = list(data.keys())
        images_sparse = []
        for obj_n in obj_names:
            images_sparse += map(lambda x: x + '.jpg',
                                 list(data[obj_n]["frames"]))
        images_sparse = np.sort(np.unique(images_sparse))

        images = np.sort(
            list(os.listdir(os.path.join(self.image_root, seq_name))))
        start_img = images_sparse[0]
        end_img = images_sparse[-1]
        for start_idx in range(len(images)):
            if start_img in images[start_idx]:
                break
        for end_idx in range(len(images))[::-1]:
            if end_img in images[end_idx]:
                break
        images = images[start_idx:(end_idx + 1)]
        labels = np.sort(
            list(os.listdir(os.path.join(self.label_root, seq_name))))

        try:
            if not os.path.isfile(
                    os.path.join(self.result_root, seq_name, labels[0])):
                if not os.path.exists(os.path.join(self.result_root,
                                                   seq_name)):
                    os.makedirs(os.path.join(self.result_root, seq_name))
                shutil.copy(
                    os.path.join(self.label_root, seq_name, labels[0]),
                    os.path.join(self.result_root, seq_name, labels[0]))
        except Exception as inst:
            print(inst)
            print('Failed to create a result folder for sequence {}.'.format(
                seq_name))

        seq_dataset = VOSTest(self.image_root,
                              self.label_root,
                              seq_name,
                              images,
                              labels,
                              transform=self.transform,
                              rgb=self.rgb)
        seq_dataset.images_sparse = images_sparse

        return seq_dataset

    def _check_preprocess(self):
        _seq_list_file = self.seq_list_file
        if not os.path.isfile(_seq_list_file):
            print(_seq_list_file)
            return False
        else:
            self.ann_f = json.load(open(self.seq_list_file, 'r'))['videos']
            return True


class DAVIS_Test(object):
    def __init__(self,
                 split=['val'],
                 root='./DAVIS',
                 year=2017,
                 transform=None,
                 rgb=True,
                 full_resolution=False,
                 result_root=None):
        self.transform = transform
        self.rgb = rgb
        self.result_root = result_root
        if year == 2016:
            self.single_obj = True
        else:
            self.single_obj = False
        if full_resolution:
            resolution = 'Full-Resolution'
        else:
            resolution = '480p'
        self.image_root = os.path.join(root, 'JPEGImages', resolution)
        self.label_root = os.path.join(root, 'Annotations', resolution)
        seq_names = []
        for spt in split:
            if spt == 'test':
                spt = 'test-dev'
            with open(os.path.join(root, 'ImageSets', str(year),
                                   spt + '.txt')) as f:
                seqs_tmp = f.readlines()
            seqs_tmp = list(map(lambda elem: elem.strip(), seqs_tmp))
            seq_names.extend(seqs_tmp)
        self.seqs = list(np.unique(seq_names))

    def __len__(self):
        return len(self.seqs)

    def __getitem__(self, idx):
        seq_name = self.seqs[idx]
        images = list(
            np.sort(os.listdir(os.path.join(self.image_root, seq_name))))
        labels = [images[0].replace('jpg', 'png')]

        if not os.path.isfile(
                os.path.join(self.result_root, seq_name, labels[0])):
            seq_result_folder = os.path.join(self.result_root, seq_name)
            try:
                if not os.path.exists(seq_result_folder):
                    os.makedirs(seq_result_folder)
            except Exception as inst:
                print(inst)
                print(
                    'Failed to create a result folder for sequence {}.'.format(
                        seq_name))
            source_label_path = os.path.join(self.label_root, seq_name,
                                             labels[0])
            result_label_path = os.path.join(self.result_root, seq_name,
                                             labels[0])
            if self.single_obj:
                label = Image.open(source_label_path)
                label = np.array(label, dtype=np.uint8)
                label = (label > 0).astype(np.uint8)
                label = Image.fromarray(label).convert('P')
                label.putpalette(_palette)
                label.save(result_label_path)
            else:
                shutil.copy(source_label_path, result_label_path)

        seq_dataset = VOSTest(self.image_root,
                              self.label_root,
                              seq_name,
                              images,
                              labels,
                              transform=self.transform,
                              rgb=self.rgb,
                              single_obj=self.single_obj,
                              resolution=480)
        return seq_dataset


class _EVAL_TEST(Dataset):
    def __init__(self, transform, seq_name):
        self.seq_name = seq_name
        self.num_frame = 10
        self.transform = transform

    def __len__(self):
        return self.num_frame

    def __getitem__(self, idx):
        current_frame_obj_num = 2
        height = 400
        width = 400
        img_name = 'test{}.jpg'.format(idx)
        current_img = np.zeros((height, width, 3)).astype(np.float32)
        if idx == 0:
            current_label = (current_frame_obj_num * np.ones(
                (height, width))).astype(np.uint8)
            sample = {
                'current_img': current_img,
                'current_label': current_label
            }
        else:
            sample = {'current_img': current_img}

        sample['meta'] = {
            'seq_name': self.seq_name,
            'frame_num': self.num_frame,
            'obj_num': current_frame_obj_num,
            'current_name': img_name,
            'height': height,
            'width': width,
            'flip': False
        }

        if self.transform is not None:
            sample = self.transform(sample)
        return sample


class EVAL_TEST(object):
    def __init__(self, transform=None, result_root=None):
        self.transform = transform
        self.result_root = result_root

        self.seqs = ['test1', 'test2', 'test3']

    def __len__(self):
        return len(self.seqs)

    def __getitem__(self, idx):
        seq_name = self.seqs[idx]

        if not os.path.exists(os.path.join(self.result_root, seq_name)):
            os.makedirs(os.path.join(self.result_root, seq_name))

        seq_dataset = _EVAL_TEST(self.transform, seq_name)
        return seq_dataset
